"""DVF devices."""

import numpy as _np

from mathphys.functions import get_namedtuple as _get_namedtuple

from .device import DeviceNC as _DeviceNC


class DVF(_DeviceNC):
    """Beam Visualization Device ("Dispositivo de Visualização de Feixe")."""

    class DEVICES:
        """Devices names."""
        CAX_DVF1 = 'CAX:A:BASLER01'
        CAX_DVF2 = 'CAX:B:BASLER01'
        ALL = (CAX_DVF1, CAX_DVF2)

    _default_timeout = 10  # [s]

    _dvfparam_fields = (
        'ACQUISITION_TIME_MIN',  # [s]
        'ACQUISITION_TIME_DEFAULT',  # [s]
        'EXPOSURE_TIME_DEFAULT',  # [s]
        'IMAGE_SIZE_V',  # [pixel]
        'IMAGE_SIZE_H',  # [pixel]
        'IMAGE_PIXEL_SIZE',  # [um]
        'OPTICS_MAGNIFICATION_FACTOR',  # source to image
    )
    #   DVF device :       ((sizey, sizex), pixel_size_um mag_factor
    _dev2params = {
        DEVICES.CAX_DVF1 :
            # DVF1 Today: pixel size 4.8 um; magnification factor 0.5
            _get_namedtuple('DVFParameters',
            _dvfparam_fields, (0.5, 0.5, 0.005, 1024, 1280, 4.8, 0.5)),
        DEVICES.CAX_DVF2 :
            # DVF2 today: pixel size 4.8 um; magnification factor 5.0
            # DVF2 future hifi: pixel size 2.4 um; magnification factor 5.0
            _get_namedtuple('DVFParameters',
            _dvfparam_fields, (0.5, 0.5, 0.005, 1024, 1280, 4.8, 5.0)),
        }

    _properties = (
        'ffmstream1:EnableCallbacks', 'ffmstream1:EnableCallbacks_RBV',
        'Trans1:EnableCallbacks', 'Trans1:EnableCallbacks_RBV',
        'cam1:ArrayCallbacks', 'cam1:ArrayCallbacks_RBV',
        'cam1:AcquireTime', 'cam1:AcquireTime_RBV',
        'cam1:AcquirePeriod', 'cam1:AcquirePeriod_RBV',
        'cam1:Acquire',
        'image1:EnableCallbacks', 'image1:EnableCallbacks_RBV',
        'image1:ArrayData',
        )

    def __init__(self, devname):
        """Init."""
        # check if device exists
        if devname not in DVF.DEVICES.ALL:
            raise NotImplementedError(devname)
        # call base class constructor
        super().__init__(devname, properties=DVF._properties)

    def parameters(self):
        """Return DVF parameters."""
        return self._dev2params[self.devname]

    @property
    def exposure_time(self):
        """Camera exposure time [s]."""
        return self['cam1:AcquireTime']

    @exposure_time.setter
    def exposure_time(self, value):
        """Camera exposure time [s]."""
        value = max(value, 0)
        self['cam1:AcquireTime'] = value

    @property
    def acquisition_time(self):
        """Camera acquisition time [s]."""
        return self['cam1:AcquirePeriod']

    @acquisition_time.setter
    def acquisition_time(self, value):
        """Camera acquisition time [s]."""
        params = self.parameters
        value = max(value, params.ACQUISITION_TIME_MIN)
        self['cam1:AcquirePeriod'] = value

    @property
    def acquisition_status(self):
        """Return IOC acquisition status."""
        return self['cam1:Acquire']

    @property
    def image_sizeh(self):
        """Image horizontal size (pixels)."""
        params = self.parameters
        return params.IMAGE_SIZE_H

    @property
    def image_sizev(self):
        """Image vertical size (pixels)."""
        params = self.parameters
        return params.IMAGE_SIZE_V

    @property
    def image(self):
        """Return DVF image formatted as a (sizey, sizex) matriz."""
        params = self.parameters
        shape = (params.IMAGE_SIZE_V, params.IMAGE_SIZE_H)
        data = self['image1:ArrayData']
        image = _np.reshape(data, shape)
        return image

    @property
    def image_pixel_size(self):
        """Image pixel size [um]."""
        params = self.parameters
        return params.IMAGE_PIXEL_SIZE

    @property
    def optics_magnefication_factor(self):
        """Source to image magnefication factor."""
        params = self.parameters
        return params.OPTICS_MAGNIFICATION_FACTOR

    @property
    def conv_pixel_2_srcsize(self):
        """Pixel to source size convertion factor."""
        pixel_size = self.image_pixel_size
        mag_factor = self.optics_magnefication_factor
        pixel2srcsize = pixel_size / mag_factor
        return pixel2srcsize

    def cmd_reset(self, timeout=None):
        """Reset DVF to a standard configuration."""
        props_values = {
            'cam1:EnableCallbacks': 1,  # Enable
            'image1:EnableCallbacks': 1,  # Enable
            'ffmstream1:EnableCallbacks': 1,  # Enable
            'Trans1:EnableCallbacks': 1,  # Enable    
        }
        for propty, value in props_values.items():
            self[propty] = value
        params = self.parameters
        self.exposure_time = params.EXPOSURE_TIME_DEFAULT
        self.acquisition_time = params.ACQUISITION_TIME_DEFAULT
        for propty, value in props_values.items():
            self._wait(propty, value, timeout=timeout, comp='eq')

    def cmd_acquire_on(self, timeout=None):
        """Tune IOC image acquisition on."""
        self._set_and_wait('cam1:Acquire', 1, timeout=timeout)

    def cmd_acquire_off(self, timeout=None):
        """Tune IOC image acquisition off."""
        self._set_and_wait('cam1:Acquire', 0, timeout=timeout)

    def _set_and_wait(self, propty, value, timeout=None):
        """."""
        timeout = timeout or self._default_timeout
        self[propty] = value
        self._wait(propty + '_RBV', value, timeout=timeout, comp='eq')




"""Orbit Interlock devices."""

import numpy as _np

from ..search import BPMSearch as _BPMSearch
from ..util import ClassProperty as _classproperty
from .device import Device as _Device, Devices as _Devices


class BaseOrbitIntlk:
    """Base BPM Orbit Interlock Info."""

    CONV_NM2UM = 1e-3  # [nm] --> [um]

    UP_2_DOWN = {
        'M1': 'M2',
        'C1-1': 'C1-2',
        'C2': 'C3-1',
        'C3-2': 'C4',
    }
    DOWN_2_UP = {val: key for key, val in UP_2_DOWN.items()}

    __BPM_NAMES = None
    __BPM_NICKNAMES = None
    __BPM_POS = None
    __BPM_DOWNSTREAM = None
    __BPM_UPSTREAM = None

    @_classproperty
    def BPM_NAMES(cls):
        """BPM names."""
        if cls.__BPM_NAMES is not None:
            return cls.__BPM_NAMES

        cls.__BPM_NAMES = _BPMSearch.get_names({'sec': 'SI', 'dev': 'BPM'})
        cls.__BPM_NICKNAMES = _BPMSearch.get_nicknames(cls.__BPM_NAMES)
        cls.__BPM_POS = _BPMSearch.get_positions(cls.__BPM_NAMES)

        def _parse_nick_down_up(name, nick):
            sub, idx, *_ = nick.split('-') + ['']
            return name.substitute(sub=name.sub[:2]+sub, idx=idx)

        downnames, upnames = [], []
        for bpm in cls.__BPM_NAMES:
            nick = bpm.sub[2:]+('-' + bpm.idx if bpm.idx else '')
            if nick in BaseOrbitIntlk.DOWN_2_UP:
                down = bpm
                upnick = BaseOrbitIntlk.DOWN_2_UP[nick]
                upn = _parse_nick_down_up(bpm, upnick)
            elif nick in BaseOrbitIntlk.UP_2_DOWN:
                upn = bpm
                downnick = BaseOrbitIntlk.UP_2_DOWN[nick]
                down = _parse_nick_down_up(bpm, downnick)
            else:
                down, upn = '', ''
            downnames.append(down)
            upnames.append(upn)
        cls.__BPM_DOWNSTREAM = downnames
        cls.__BPM_UPSTREAM = upnames
        return cls.__BPM_NAMES

    @_classproperty
    def BPM_NICKNAMES(cls):
        """BPM nicknames."""
        cls.BPM_NAMES
        return cls.__BPM_NICKNAMES

    @_classproperty
    def BPM_POS(cls):
        """BPM positions."""
        cls.BPM_NAMES
        return cls.__BPM_POS

    @_classproperty
    def BPM_DOWNSTREAM(cls):
        """BPM downstream."""
        cls.BPM_NAMES
        return cls.__BPM_DOWNSTREAM

    @_classproperty
    def BPM_UPSTREAM(cls):
        """BPM upstream."""
        cls.BPM_NAMES
        return cls.__BPM_UPSTREAM

    def __init__(self):
        """Init."""
        self._oper = {
            'mean': self._mean,
            'diff': self._diff,
        }

    @staticmethod
    def get_down_up_bpms(bpmname):
        """Return down and up BPM names for bpmname."""
        index = BaseOrbitIntlk.BPM_NAMES.index(bpmname)
        down = BaseOrbitIntlk.BPM_DOWNSTREAM[index]
        upn = BaseOrbitIntlk.BPM_UPSTREAM[index]
        return down, upn

    def calc_intlk_metric(self, posarray, operation='', metric=''):
        """Return interlock metric, translation or angulation."""
        if not operation:
            if not metric:
                raise ValueError(
                    'either the operation or the metric is required')
            operation = 'mean' if 'trans' in metric.lower() else 'diff'

        data_values = list()
        for bpm in BaseOrbitIntlk.BPM_NAMES:
            down, upn = BaseOrbitIntlk.get_down_up_bpms(bpm)
            dval = posarray[BaseOrbitIntlk.BPM_NAMES.index(down)]
            uval = posarray[BaseOrbitIntlk.BPM_NAMES.index(upn)]
            func = self._oper[operation]
            val = func(dval, uval)
            data_values.append(val)
        return data_values

    @staticmethod
    def _mean(var1, var2):
        return (var1 + var2)/2

    @staticmethod
    def _diff(var1, var2):
        return var1 - var2


class BPMOrbitIntlk(BaseOrbitIntlk, _Device):
    """This device group the orbit interlock  PVs from one BPM."""

    _properties = (
        # ==============================================================
        # Basic properties
        'PosX-Mon', 'PosY-Mon', 'Sum-Mon',
        # ==============================================================
        # General
        # +++++++
        # General interlock enable:
        'IntlkEn-Sel', 'IntlkEn-Sts',
        # General interlock clear:
        'IntlkClr-Sel',  # maybe -Cmd?
        # Minimum sum threshold enable:
        # Habilita interlock de órbita apenas quando threshold da soma
        # ultrapassar o valor em "IntlkLmtMinSum-SP"
        'IntlkMinSumEn-Sel', 'IntlkMinSumEn-Sts',
        # Minimum sum threshold (em contagens da Soma da taxa Monit1):
        'IntlkLmtMinSum-SP', 'IntlkLmtMinSum-RB',
        # Status Instantâneo:
        # Interlock instântaneo, dificilmente será detectado com
        # a implementação atual do gateware
        'Intlk-Mon',
        # Latch do interlock, limpo apenas acionando-se a PV "Clr"
        # correspondente
        'IntlkLtc-Mon',
        # ===============================================================
        # Translation (interlock de translação)
        # +++++++++++++++++++++++++++++++++++++
        # ***************************************************************
        # Condição para interlock de translação:
        #   thres_min > (pos BPM downstream + pos BPM upstream)/2 or
        #   thres_max < (pos BPM downstream + pos BPM upstream)/2
        # BPMs são agrupados 2 a 2 seguindo a ordem do feixe:
        # - M1/M2
        # - C1-1/C1-2
        # - C2/C3-1
        # - C3-2/C4
        # BPM upstream é sempre o "primeiro" BPM da dupla acima e BPM
        # downstream é sempre o "segundo" BPM da dupla.
        # ***************************************************************
        # Translation interlock enable:
        'IntlkTransEn-Sel', 'IntlkTransEn-Sts',
        # Translation interlock clear:
        'IntlkTransClr-Sel',
        # Thresholds (em nm da taxa Monit1):
        'IntlkLmtTransMaxX-SP', 'IntlkLmtTransMaxX-RB',
        'IntlkLmtTransMinX-SP', 'IntlkLmtTransMinX-RB',
        'IntlkLmtTransMaxY-SP', 'IntlkLmtTransMaxY-RB',
        'IntlkLmtTransMinY-SP', 'IntlkLmtTransMinY-RB',
        # Status Instantâneo:
        #   XouY mascarado pelo "Enable"
        'IntlkTransSmaller-Mon', 'IntlkTransBigger-Mon',
        # the ones bellow are not masked by "Enable"
        'IntlkTransSmallerAny-Mon', 'IntlkTransBiggerAny-Mon',  # X ou Y
        'IntlkTransSmallerX-Mon', 'IntlkTransBiggerX-Mon',  # X
        'IntlkTransSmallerY-Mon', 'IntlkTransBiggerY-Mon',  # Y
        # limpo apenas acionando-se a PV "Clr" correspondente
        'IntlkTransSmallerLtc-Mon', 'IntlkTransBiggerLtc-Mon',
        'IntlkTransSmallerLtcX-Mon', 'IntlkTransBiggerLtcX-Mon',
        'IntlkTransSmallerLtcY-Mon', 'IntlkTransBiggerLtcY-Mon',
        # =============================================================
        # Angular (interlock de ângulo)
        # +++++++++++++++++++++++++++++
        # *************************************************************
        # Condição para interlock de ângulo:
        #   thres_min > (posição BPM downstream - posição BPM upstream) or
        #   thres_max < (posição BPM downstream - posição BPM upstream)
        # BPMs são agrupados 2 a 2 seguindo a ordem do feixe:
        # - M1/M2
        # - C1-1/C1-2
        # - C2/C3-1
        # - C3-2/C4
        # BPM upstream é sempre o "primeiro" BPM da dupla acima e BPM
        # downstream é sempre o "segundo" BPM da dupla.
        # ************************************************************
        # Angulation interlock enable:
        'IntlkAngEn-Sel', 'IntlkAngEn-Sts',
        # Angulation interlock clear:
        'IntlkAngClr-Sel',
        # Thresholds (em rad.nm da taxa Monit1).
        #  Thresholds devem ser calculados como ângulo (em rad)
        #  entre os 2 BPMs adjacentes * distância (em nm) entre eles):
        'IntlkLmtAngMaxX-SP', 'IntlkLmtAngMaxX-RB',
        'IntlkLmtAngMinX-SP', 'IntlkLmtAngMinX-RB',
        'IntlkLmtAngMaxY-SP', 'IntlkLmtAngMaxY-RB',
        'IntlkLmtAngMinY-SP', 'IntlkLmtAngMinY-RB',
        # Status Instantâneo:
        #  X ou Y mascarado pelo "Enable"
        'IntlkAngSmaller-Mon', 'IntlkAngBigger-Mon',
        # the ones bellow are not masked by "Enable"
        'IntlkAngSmallerAny-Mon', 'IntlkAngBiggerAny-Mon',  # X ou Y
        'IntlkAngSmallerX-Mon', 'IntlkAngBiggerX-Mon',  # X
        'IntlkAngSmallerY-Mon', 'IntlkAngBiggerY-Mon',  # Y
        # limpo apenas acionando-se a PV "Clr" correspondente
        'IntlkAngSmallerLtc-Mon', 'IntlkAngBiggerLtc-Mon',
        'IntlkAngSmallerLtcX-Mon', 'IntlkAngBiggerLtcX-Mon',
        'IntlkAngSmallerLtcY-Mon', 'IntlkAngBiggerLtcY-Mon',
        # ============================================================
        )

    def __init__(self, devname):
        """Init."""
        # call base class constructor
        BaseOrbitIntlk.__init__(self)
        if devname not in self.BPM_NAMES:
            raise ValueError(devname + ' is no a valid BPM name.')
        _Device.__init__(self, devname, properties=BPMOrbitIntlk._properties)

    @property
    def posx(self):
        """Position X, Monit rate."""
        return self['PosX-Mon'] * self.CONV_NM2UM

    @property
    def posy(self):
        """Position Y, Monit rate."""
        return self['PosY-Mon'] * self.CONV_NM2UM

    @property
    def possum(self):
        """Sum, Monit rate."""
        return self['Sum-Mon']

    @property
    def pair_down_up_bpms(self):
        """Return pair of downstream and upstream BPM names."""
        return self.get_down_up_bpms(self.devname)

    # --- general interlock ---

    @property
    def gen_enable(self):
        """General interlock enable."""
        return self['IntlkEn-Sts']

    @gen_enable.setter
    def gen_enable(self, value):
        self['IntlkEn-Sel'] = int(value)

    def cmd_reset_gen(self):
        """General interlock clear."""
        self['IntlkClr-Sel'] = 1
        return True

    @property
    def gen_inst(self):
        """Instantaneous general interlock."""
        return self['Intlk-Mon']

    @property
    def gen_latch(self):
        """Latch general interlock."""
        return self['IntlkLtc-Mon']

    # --- minimum sum threshold ---

    @property
    def minsumthres_enable(self):
        """
        Minimum sum threshold enable.

        If enabled, generate orbit interlock only when sum threshold exceeds
        value in 'minsumthres' property.
        """
        return self['IntlkMinSumEn-Sts']

    @minsumthres_enable.setter
    def minsumthres_enable(self, value):
        self['IntlkMinSumEn-Sel'] = int(value)

    @property
    def minsumthres(self):
        """Minimum sum threshold [sum count, Monit1 rate]."""
        return self['IntlkLmtMinSum-RB']

    @minsumthres.setter
    def minsumthres(self, value):
        self['IntlkLmtMinSum-SP'] = int(value)

    # --- translation interlock ---

    @property
    def trans_enable(self):
        """Translation interlock enable."""
        return self['IntlkTransEn-Sts']

    @trans_enable.setter
    def trans_enable(self, value):
        self['IntlkTransEn-Sel'] = int(value)

    def cmd_reset_trans(self):
        """Translation interlock clear."""
        self['IntlkTransClr-Sel'] = 1
        return True

    @property
    def trans_thresminx(self):
        """Minimum X translation threshold."""
        return self['IntlkLmtTransMinX-RB']

    @trans_thresminx.setter
    def trans_thresminx(self, value):
        self['IntlkLmtTransMinX-SP'] = value

    @property
    def trans_thresmaxx(self):
        """Maximum X translation threshold."""
        return self['IntlkLmtTransMaxX-RB']

    @trans_thresmaxx.setter
    def trans_thresmaxx(self, value):
        self['IntlkLmtTransMaxX-SP'] = value

    @property
    def trans_thresminy(self):
        """Minimum Y translation threshold."""
        return self['IntlkLmtTransMinY-RB']

    @trans_thresminy.setter
    def trans_thresminy(self, value):
        self['IntlkLmtTransMinY-SP'] = value

    @property
    def trans_thresmaxy(self):
        """Maximum Y translation threshold."""
        return self['IntlkLmtTransMaxY-RB']

    @trans_thresmaxy.setter
    def trans_thresmaxy(self, value):
        self['IntlkLmtTransMaxY-SP'] = value

    @property
    def trans_mask_smaller(self):
        """
        Instantaneous translation interlock set when either X or Y
        minimum thresholds are exceeded, masked by general enable.
        """
        return self['IntlkTransSmaller-Mon']

    @property
    def trans_mask_bigger(self):
        """
        Instantaneous translation interlock set when either X or Y
        maximum thresholds are exceeded, masked by general enable.
        """
        return self['IntlkTransBigger-Mon']

    @property
    def trans_inst_smaller(self):
        """
        Instantaneous translation interlock set when either X or Y
        minimum thresholds are exceeded.
        """
        return self['IntlkTransSmallerAny-Mon']

    @property
    def trans_inst_bigger(self):
        """
        Instantaneous translation interlock set when either X or Y
        maximum thresholds are exceeded.
        """
        return self['IntlkTransBiggerAny-Mon']

    @property
    def trans_inst_smaller_x(self):
        """
        Instantaneous translation interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkTransSmallerX-Mon']

    @property
    def trans_inst_bigger_x(self):
        """
        Instantaneous translation interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkTransBiggerX-Mon']

    @property
    def trans_inst_smaller_y(self):
        """
        Instantaneous translation interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkTransSmallerY-Mon']

    @property
    def trans_inst_bigger_y(self):
        """
        Instantaneous translation interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkTransBiggerY-Mon']

    @property
    def trans_latch_smaller(self):
        """
        Latch translation interlock set when either X or Y
        minimum thresholds are exceeded.
        """
        return self['IntlkTransSmallerLtc-Mon']

    @property
    def trans_latch_bigger(self):
        """
        Latch translation interlock set when either X or Y
        maximum thresholds are exceeded.
        """
        return self['IntlkTransBiggerLtc-Mon']

    @property
    def trans_latch_smaller_x(self):
        """
        Latch translation interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkTransSmallerLtcX-Mon']

    @property
    def trans_latch_bigger_x(self):
        """
        Latch translation interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkTransBiggerLtcX-Mon']

    @property
    def trans_latch_smaller_y(self):
        """
        Latch translation interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkTransSmallerLtcY-Mon']

    @property
    def trans_latch_bigger_y(self):
        """
        Latch translation interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkTransBiggerLtcY-Mon']

    #  --- angulation interlock ---

    @property
    def ang_enable(self):
        """Angulation interlock enable."""
        return self['IntlkAngEn-Sts']

    @ang_enable.setter
    def ang_enable(self, value):
        self['IntlkAngEn-Sel'] = int(value)

    def cmd_reset_ang(self):
        """Angulation interlock clear."""
        self['IntlkAngClr-Sel'] = 1
        return True

    @property
    def ang_thresminx(self):
        """Minimum X angulation threshold."""
        return self['IntlkLmtAngMinX-RB']

    @ang_thresminx.setter
    def ang_thresminx(self, value):
        self['IntlkLmtAngMinX-SP'] = value

    @property
    def ang_thresmaxx(self):
        """Maximum X angulation threshold."""
        return self['IntlkLmtAngMaxX-RB']

    @ang_thresmaxx.setter
    def ang_thresmaxx(self, value):
        self['IntlkLmtAngMaxX-SP'] = value

    @property
    def ang_thresminy(self):
        """Minimum Y angulation threshold."""
        return self['IntlkLmtAngMinY-RB']

    @ang_thresminy.setter
    def ang_thresminy(self, value):
        self['IntlkLmtAngMinY-SP'] = value

    @property
    def ang_thresmaxy(self):
        """Maximum Y angulation threshold."""
        return self['IntlkLmtAngMaxY-RB']

    @ang_thresmaxy.setter
    def ang_thresmaxy(self, value):
        self['IntlkLmtAngMaxY-SP'] = value

    @property
    def ang_mask_smaller(self):
        """
        Instantaneous angulation interlock set when either X or Y
        minimum thresholds are exceeded, masked by general enable.
        """
        return self['IntlkAngSmaller-Mon']

    @property
    def ang_mask_bigger(self):
        """
        Instantaneous angulation interlock set when either X or Y
        maximum thresholds are exceeded, masked by general enable.
        """
        return self['IntlkAngBigger-Mon']

    @property
    def ang_inst_smaller(self):
        """
        Instantaneous angulation interlock set when either X or Y
        minimum thresholds are exceeded.
        """
        return self['IntlkAngSmallerAny-Mon']

    @property
    def ang_inst_bigger(self):
        """
        Instantaneous angulation interlock set when either X or Y
        maximum thresholds are exceeded.
        """
        return self['IntlkAngBiggerAny-Mon']

    @property
    def ang_inst_smaller_x(self):
        """
        Instantaneous angulation interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkAngSmallerX-Mon']

    @property
    def ang_inst_bigger_x(self):
        """
        Instantaneous angulation interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkAngBiggerX-Mon']

    @property
    def ang_inst_smaller_y(self):
        """
        Instantaneous angulation interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkAngSmallerY-Mon']

    @property
    def ang_inst_bigger_y(self):
        """
        Instantaneous angulation interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkAngBiggerY-Mon']

    @property
    def ang_latch_smaller(self):
        """
        Latch angulation interlock set when either X or Y
        minimum thresholds are exceeded.
        """
        return self['IntlkAngSmallerLtc-Mon']

    @property
    def ang_latch_bigger(self):
        """
        Latch angulation interlock set when either X or Y
        maximum thresholds are exceeded.
        """
        return self['IntlkAngBiggerLtc-Mon']

    @property
    def ang_latch_smaller_x(self):
        """
        Latch angulation interlock set when X
        minimum threshold is exceeded.
        """
        return self['IntlkAngSmallerLtcX-Mon']

    @property
    def ang_latch_bigger_x(self):
        """
        Latch angulation interlock set when X
        maximum threshold is exceeded.
        """
        return self['IntlkAngBiggerLtcX-Mon']

    @property
    def ang_latch_smaller_y(self):
        """
        Latch angulation interlock set when Y
        minimum threshold is exceeded.
        """
        return self['IntlkAngSmallerLtcY-Mon']

    @property
    def ang_latch_bigger_y(self):
        """
        Latch angulation interlock set when Y
        maximum threshold is exceeded.
        """
        return self['IntlkAngBiggerLtcY-Mon']


class OrbitInterlock(BaseOrbitIntlk, _Devices):
    """Orbit Interlock device."""

    TIMEOUT = 10

    # NOTE: this is a temporary location, will migrate to csconstants.
    # The values will be replaced by the measured calibration data.
    _av = _np.array([
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0, 1.0,
        1.0, 1.0, 1.0, 1.0, 1.0, 1.0])
    _bv = _np.array([
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0, 0.0,
        0.0, 0.0, 0.0, 0.0, 0.0, 0.0])

    # for each BPM, folowing BPM_NAMES indices
    # a, b: monit = a*monit1 + b
    CONV_POLY_MONIT1_2_MONIT = _np.vstack([_av, _bv])

    class DEVICES:
        """."""
        SI = 'SI-Fam:DI-BPM'
        ALL = (SI, )

    def __init__(self, devname=None):
        """Init."""
        if devname is None:
            devname = self.DEVICES.SI
        if devname not in self.DEVICES.ALL:
            raise ValueError('Wrong value for devname')
        BaseOrbitIntlk.__init__(self)
        devs = [BPMOrbitIntlk(dev) for dev in self.BPM_NAMES]
        _Devices.__init__(self, devname, devs)

    # --- general interlock ---

    def cmd_gen_enable(self, timeout=TIMEOUT):
        """Enable all BPM general interlock."""
        for dev in self.devices:
            dev.gen_enable = 1
        return self._wait_devices_propty(
            self.devices, 'IntlkEn-Sts', 1, timeout=timeout)

    def cmd_gen_disable(self, timeout=TIMEOUT):
        """Disable all BPM general interlock."""
        for dev in self.devices:
            dev.gen_enable = 0
        return self._wait_devices_propty(
            self.devices, 'IntlkEn-Sts', 0, timeout=timeout)

    def cmd_reset_gen(self):
        """Reset all BPM general interlock."""
        for dev in self.devices:
            dev.cmd_reset_gen()
        return True

    @property
    def gen_inst(self):
        """General instantaneous interlocks.

        Returns:
            gen_inst (numpy.ndarray, 160):
                general instantaneous interlock for each BPM.
        """
        return _np.array([b.gen_inst for b in self._devices])

    @property
    def gen_latch(self):
        """General latch interlocks.

        Returns:
            gen_latch (numpy.ndarray, 160):
                general latch interlock for each BPM.
        """
        return _np.array([b.gen_latch for b in self._devices])

    # --- minimum sum threshold ---

    def cmd_minsumthres_enable(self, timeout=TIMEOUT):
        """Enable all BPM minimum sum threshold."""
        for dev in self.devices:
            dev.minsumthres_enable = 1
        return self._wait_devices_propty(
            self.devices, 'IntlkMinSumEn-Sts', 1, timeout=timeout)

    def cmd_minsumthres_disable(self, timeout=TIMEOUT):
        """Disable all BPM minimum sum threshold."""
        for dev in self.devices:
            dev.minsumthres_enable = 0
        return self._wait_devices_propty(
            self.devices, 'IntlkMinSumEn-Sts', 0, timeout=timeout)

    @property
    def minsumthres(self):
        """Minimum sum thresholds.

        Returns:
            thres (numpy.ndarray, 160): min.sum threshold for each BPM.
        """
        return _np.array([b.minsumthres for b in self._devices])

    @minsumthres.setter
    def minsumthres(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.minsumthres = value[idx]

    # --- translation interlock ---

    def cmd_trans_enable(self, timeout=TIMEOUT):
        """Enable all BPM translation interlock."""
        for dev in self.devices:
            dev.trans_enable = 1
        return self._wait_devices_propty(
            self.devices, 'IntlkTransEn-Sts', 1, timeout=timeout)

    def cmd_trans_disable(self, timeout=TIMEOUT):
        """Disable all BPM translation interlock."""
        for dev in self.devices:
            dev.trans_enable = 0
        return self._wait_devices_propty(
            self.devices, 'IntlkTransEn-Sts', 0, timeout=timeout)

    def cmd_reset_trans(self):
        """Reset all BPM translation interlock."""
        for dev in self.devices:
            dev.cmd_reset_trans()
        return True

    @property
    def trans_thresminx(self):
        """Minimum x translation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. x translation threshold for each BPM.
        """
        return _np.array([b.trans_thresminx for b in self._devices])

    @trans_thresminx.setter
    def trans_thresminx(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.trans_thresminx = value[idx]

    @property
    def trans_thresmaxx(self):
        """Maximum x translation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. x translation threshold for each BPM.
        """
        return _np.array([b.trans_thresmaxx for b in self._devices])

    @trans_thresmaxx.setter
    def trans_thresmaxx(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.trans_thresmaxx = value[idx]

    @property
    def trans_thresminy(self):
        """Minimum y translation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. y translation threshold for each BPM.
        """
        return _np.array([b.trans_thresminy for b in self._devices])

    @trans_thresminy.setter
    def trans_thresminy(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.trans_thresminy = value[idx]

    @property
    def trans_thresmaxy(self):
        """Maximum y translation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. y translation threshold for each BPM.
        """
        return _np.array([b.trans_thresmaxy for b in self._devices])

    @trans_thresmaxy.setter
    def trans_thresmaxy(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.trans_thresmaxy = value[idx]

    @property
    def trans_mask_smaller(self):
        """Instantaneous translation interlock set when either X or Y
        minimum thresholds are exceeded, masked by general enable.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_mask_smaller for b in self._devices])

    @property
    def trans_mask_bigger(self):
        """Instantaneous translation interlock set when either X or Y
        maximum thresholds are exceeded, masked by general enable.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_mask_bigger for b in self._devices])

    @property
    def trans_inst_smaller(self):
        """Instantaneous translation interlock set when either X or Y
        minimum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_inst_smaller for b in self._devices])

    @property
    def trans_inst_bigger(self):
        """Instantaneous translation interlock set when either X or Y
        maximum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_inst_bigger for b in self._devices])

    @property
    def trans_inst_smaller_x(self):
        """Instantaneous translation interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_inst_smaller_x for b in self._devices])

    @property
    def trans_inst_bigger_x(self):
        """Instantaneous translation interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_inst_bigger_x for b in self._devices])

    @property
    def trans_inst_smaller_y(self):
        """Instantaneous translation interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_inst_smaller_y for b in self._devices])

    @property
    def trans_inst_bigger_y(self):
        """Instantaneous translation interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_inst_bigger_y for b in self._devices])

    @property
    def trans_latch_smaller(self):
        """Latch translation interlock set when either X or Y
        minimum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_latch_smaller for b in self._devices])

    @property
    def trans_latch_bigger(self):
        """Latch translation interlock set when either X or Y
        maximum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_latch_bigger for b in self._devices])

    @property
    def trans_latch_smaller_x(self):
        """Latch translation interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_latch_smaller_x for b in self._devices])

    @property
    def trans_latch_bigger_x(self):
        """Latch translation interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_latch_bigger_x for b in self._devices])

    @property
    def trans_latch_smaller_y(self):
        """Latch translation interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_latch_smaller_y for b in self._devices])

    @property
    def trans_latch_bigger_y(self):
        """Latch translation interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.trans_latch_bigger_y for b in self._devices])

    #  --- angulation interlock ---

    def cmd_ang_enable(self, timeout=TIMEOUT):
        """Enable all BPM angulation interlock."""
        for dev in self.devices:
            dev.ang_enable = 1
        return self._wait_devices_propty(
            self.devices, 'IntlkAngEn-Sts', 1, timeout=timeout)

    def cmd_ang_disable(self, timeout=TIMEOUT):
        """Disable all BPM angulation interlock."""
        for dev in self.devices:
            dev.ang_enable = 0
        return self._wait_devices_propty(
            self.devices, 'IntlkAngEn-Sts', 0, timeout=timeout)

    def cmd_reset_ang(self):
        """Reset all BPM angulation interlock."""
        for dev in self.devices:
            dev.cmd_reset_ang()
        return True

    @property
    def ang_thresminx(self):
        """Minimum x angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. x angulation threshold for each BPM.
        """
        return _np.array([b.ang_thresminx for b in self._devices])

    @ang_thresminx.setter
    def ang_thresminx(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.ang_thresminx = value[idx]

    @property
    def ang_thresmaxx(self):
        """Maximum x angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. x angulation threshold for each BPM.
        """
        return _np.array([b.ang_thresmaxx for b in self._devices])

    @ang_thresmaxx.setter
    def ang_thresmaxx(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.ang_thresmaxx = value[idx]

    @property
    def ang_thresminy(self):
        """Minimum y angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                min. y angulation threshold for each BPM.
        """
        return _np.array([b.ang_thresminy for b in self._devices])

    @ang_thresminy.setter
    def ang_thresminy(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.ang_thresminy = value[idx]

    @property
    def ang_thresmaxy(self):
        """Maximum y angulation thresholds.

        Returns:
            thres (numpy.ndarray, 160):
                max. y angulation threshold for each BPM.
        """
        return _np.array([b.ang_thresmaxy for b in self._devices])

    @ang_thresmaxy.setter
    def ang_thresmaxy(self, value):
        value = self._handle_thres_input(value)
        for idx, dev in enumerate(self.devices):
            dev.ang_thresmaxy = value[idx]

    @property
    def ang_mask_smaller(self):
        """Instantaneous angulation interlock set when either X or Y
        minimum thresholds are exceeded, masked by general enable.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_mask_smaller for b in self._devices])

    @property
    def ang_mask_bigger(self):
        """Instantaneous angulation interlock set when either X or Y
        maximum thresholds are exceeded, masked by general enable.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_mask_bigger for b in self._devices])

    @property
    def ang_inst_smaller(self):
        """Instantaneous angulation interlock set when either X or Y
        minimum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_smaller for b in self._devices])

    @property
    def ang_inst_bigger(self):
        """Instantaneous angulation interlock set when either X or Y
        maximum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_bigger for b in self._devices])

    @property
    def ang_inst_smaller_x(self):
        """Instantaneous angulation interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_smaller_x for b in self._devices])

    @property
    def ang_inst_bigger_x(self):
        """Instantaneous angulation interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_bigger_x for b in self._devices])

    @property
    def ang_inst_smaller_y(self):
        """Instantaneous angulation interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_smaller_y for b in self._devices])

    @property
    def ang_inst_bigger_y(self):
        """Instantaneous angulation interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_inst_bigger_y for b in self._devices])

    @property
    def ang_latch_smaller(self):
        """Latch angulation interlock set when either X or Y
        minimum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_smaller for b in self._devices])

    @property
    def ang_latch_bigger(self):
        """Latch angulation interlock set when either X or Y
        maximum thresholds are exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_bigger for b in self._devices])

    @property
    def ang_latch_smaller_x(self):
        """Latch angulation interlock set when X
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_smaller_x for b in self._devices])

    @property
    def ang_latch_bigger_x(self):
        """Latch angulation interlock set when X
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_bigger_x for b in self._devices])

    @property
    def ang_latch_smaller_y(self):
        """Latch angulation interlock set when Y
        minimum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_smaller_y for b in self._devices])

    @property
    def ang_latch_bigger_y(self):
        """Latch angulation interlock set when Y
        maximum threshold is exceeded.

        Returns:
            intlk (numpy.ndarray, 160): interlock status for each BPM.
        """
        return _np.array([b.ang_latch_bigger_y for b in self._devices])

    @property
    def slow_orbit(self):
        """Slow orbit vectors.

        Returns:
            orbx (numpy.ndarray, 160): Horizontal Orbit.
            orby (numpy.ndarray, 160): Vertical Orbit.

        """
        orbx, orby = [], []
        for bpm in self._devices:
            orbx.append(bpm.posx)
            orby.append(bpm.posy)
        orbx = _np.array(orbx)
        orby = _np.array(orby)
        return orbx, orby

    @property
    def possum(self):
        """Sum vector, at Monit rate.

        Returns:
            possum (numpy.ndarray, 160): Sum vector, at Monit rate.
        """
        return _np.array([b.possum for b in self._devices])

    @property
    def translation(self):
        """Translation vectors.

        Translation at each BPM is defined as:
            (pos BPM downstream + pos BPM upstream)/2

        Returns:
            transx (numpy.ndarray, 160): Horizontal Translation.
            transy (numpy.ndarray, 160): Vertical Translation.
        """
        orbx, orby = self.slow_orbit
        transx = _np.array(self.calc_intlk_metric(orbx, metric='trans'))
        transy = _np.array(self.calc_intlk_metric(orby, metric='trans'))
        return transx, transy

    @property
    def angulation(self):
        """Angulation vectors.

        Angulation at each BPM is defined as:
            (posição BPM downstream - posição BPM upstream)

        Returns:
            angx (numpy.ndarray, 160): Horizontal Angulation.
            angy (numpy.ndarray, 160): Vertical Angulation.
        """
        orbx, orby = self.slow_orbit
        angx = _np.array(self.calc_intlk_metric(orbx, metric='ang'))
        angy = _np.array(self.calc_intlk_metric(orby, metric='ang'))
        return angx, angy

    def _handle_thres_input(self, value):
        if isinstance(value, (int, float, _np.int_, _np.float_)):
            return [value, ] * len(self.BPM_NAMES)
        if len(value) != len(self.devices):
            raise ValueError(
                'the expected length is {0}, not {1}'.format(
                    len(self.devices), len(value)))
        return value



"""Main Module of the program."""

from time import time as _time, sleep as _sleep
import logging as _log
from functools import partial as _part
from threading import Thread as _Thread
import numpy as _np

from ..epics import PV as _PV
from ..devices import HLFOFB

from .matrix import BaseMatrix as _BaseMatrix
from .orbit import BaseOrbit as _BaseOrbit
from .correctors import BaseCorrectors as _BaseCorrectors
from .base_class import BaseClass as _BaseClass, \
    compare_kicks as _compare_kicks

INTERVAL = 1


class SOFB(_BaseClass):
    """Main Class of the IOC."""

    def __init__(
            self, acc, prefix='', callback=None, orbit=None, matrix=None,
            correctors=None, tests=False):
        """Initialize Object."""
        super().__init__(acc, prefix=prefix, callback=callback)
        _log.info('Starting SOFB...')
        self._tests = tests
        self._orbit = self._correctors = self._matrix = None
        self._loop_state = self._csorb.LoopState.Open
        self._loop_freq = self._csorb.BPMsFreq
        self._loop_print_every_num_iter = 1000
        self._loop_use_pssofb = False
        self._loop_max_orb_distortion = self._csorb.DEF_MAX_ORB_DISTORTION
        zer = _np.zeros(self._csorb.nr_corrs, dtype=float)
        self._pid_errs = [zer, zer.copy(), zer.copy()]
        self._pid_gains = dict(
            ch=dict(kp=0.0, ki=0.5, kd=0.0),
            cv=dict(kp=0.0, ki=0.5, kd=0.0),
            rf=dict(kp=0.0, ki=0.5, kd=0.0))
        self._measuring_respmat = False
        self._mancorr_gain = {'ch': 1.00, 'cv': 1.00}
        self._max_kick = {'ch': 300, 'cv': 300}
        self._max_delta_kick = {'ch': 5, 'cv': 5}
        self._meas_respmat_kick = {'ch': 15, 'cv': 22.5}
        if self.isring:
            self._mancorr_gain['rf'] = 1.00
            self._max_kick['rf'] = 1e12  # a very large value
            self._max_delta_kick['rf'] = 10
            self._meas_respmat_kick['rf'] = 75
        if self.acc == 'SI':
            self.fofb = HLFOFB()
            self._download_fofb_kicks = False
            self._download_fofb_kicks_perc = 0.01
            self._update_fofb_reforb = False
            self._update_fofb_reforb_perc = 0.0
            self._donot_affect_fofb_bpms = False
            self._project_onto_fofb_nullspace = False
            self._drive_divisor = 12
            self._drive_nrcycles = 10
            self._drive_amplitude = 5
            self._drive_phase = 0
            self._drive_corr_index = 0
            self._drive_bpm_index = 0
            self._drive_type = self._csorb.DriveType.Sine
            self._drive_state = self._csorb.DriveState.Open
        self._meas_respmat_wait = 1  # seconds
        self._dtheta = None
        self._ref_corr_kicks = None
        self._thread = None
        self._havebeam_pv = _PV('SI-Glob:AP-CurrInfo:StoredEBeam-Mon')

        self.orbit = orbit
        self.correctors = correctors
        self.matrix = matrix

    def get_map2write(self):
        """Get the database of the class."""
        dbase = {
            'LoopState-Sel': self.set_auto_corr,
            'LoopFreq-SP': self.set_auto_corr_frequency,
            'LoopPrintEveryNumIters-SP': self.set_print_every_num_iters,
            'LoopPIDKpCH-SP': _part(self.set_pid_gain, 'kp', 'ch'),
            'LoopPIDKpCV-SP': _part(self.set_pid_gain, 'kp', 'cv'),
            'LoopPIDKiCH-SP': _part(self.set_pid_gain, 'ki', 'ch'),
            'LoopPIDKiCV-SP': _part(self.set_pid_gain, 'ki', 'cv'),
            'LoopPIDKdCH-SP': _part(self.set_pid_gain, 'kd', 'ch'),
            'LoopPIDKdCV-SP': _part(self.set_pid_gain, 'kd', 'cv'),
            'LoopMaxOrbDistortion-SP': self.set_max_orbit_dist,
            'MeasRespMat-Cmd': self.set_respmat_meas_state,
            'CalcDelta-Cmd': self.calc_correction,
            'ManCorrGainCH-SP': _part(self.set_mancorr_gain, 'ch'),
            'ManCorrGainCV-SP': _part(self.set_mancorr_gain, 'cv'),
            'MaxKickCH-SP': _part(self.set_max_kick, 'ch'),
            'MaxKickCV-SP': _part(self.set_max_kick, 'cv'),
            'MaxDeltaKickCH-SP': _part(self.set_max_delta_kick, 'ch'),
            'MaxDeltaKickCV-SP': _part(self.set_max_delta_kick, 'cv'),
            'DeltaKickCH-SP': _part(
                self.set_delta_kick, self._csorb.ApplyDelta.CH),
            'DeltaKickCV-SP': _part(
                self.set_delta_kick, self._csorb.ApplyDelta.CV),
            'MeasRespMatKickCH-SP': _part(self.set_respmat_kick, 'ch'),
            'MeasRespMatKickCV-SP': _part(self.set_respmat_kick, 'cv'),
            'MeasRespMatWait-SP': self.set_respmat_wait_time,
            'ApplyDelta-Cmd': self.apply_corr,
            }
        if self.isring:
            dbase['LoopPIDKpRF-SP'] = _part(self.set_pid_gain, 'kp', 'rf')
            dbase['LoopPIDKiRF-SP'] = _part(self.set_pid_gain, 'ki', 'rf')
            dbase['LoopPIDKdRF-SP'] = _part(self.set_pid_gain, 'kd', 'rf')
            dbase['ManCorrGainRF-SP'] = _part(self.set_mancorr_gain, 'rf')
            dbase['MaxDeltaKickRF-SP'] = _part(self.set_max_delta_kick, 'rf')
            dbase['DeltaKickRF-SP'] = _part(
                self.set_delta_kick, self._csorb.ApplyDelta.RF)
            dbase['MeasRespMatKickRF-SP'] = _part(self.set_respmat_kick, 'rf')
        if self.acc == 'SI':
            dbase['CorrPSSOFBEnbl-Sel'] = self.set_use_pssofb
            dbase['FOFBDownloadKicksPerc-SP'] = self.set_fofb_download_perc
            dbase['FOFBDownloadKicks-Sel'] = _part(
                self.set_fofb_interaction_props, 'downloadkicks')
            dbase['FOFBUpdateRefOrbPerc-SP'] = self.set_fofb_updatereforb_perc
            dbase['FOFBUpdateRefOrb-Sel'] = _part(
                self.set_fofb_interaction_props, 'updatereforb')
            dbase['FOFBNullSpaceProj-Sel'] = _part(
                self.set_fofb_interaction_props, 'nullspaceproj')
            dbase['FOFBZeroDistortionAtBPMs-Sel'] = _part(
                self.set_fofb_interaction_props, 'zerodistortion')
            dbase['DriveFreqDivisor-SP'] = self.set_drive_divisor
            dbase['DriveNrCycles-SP'] = self.set_drive_nrcycles
            dbase['DriveAmplitude-SP'] = self.set_drive_amplitude
            dbase['DrivePhase-SP'] = self.set_drive_phase
            dbase['DriveCorrIndex-SP'] = self.set_drive_corr_index
            dbase['DriveBPMIndex-SP'] = self.set_drive_bpm_index
            dbase['DriveType-Sel'] = self.set_drive_type
            dbase['DriveState-Sel'] = self.set_drive_state
        return dbase

    @property
    def orbit(self):
        """."""
        return self._orbit

    @orbit.setter
    def orbit(self, orb):
        """."""
        if isinstance(orb, _BaseOrbit):
            self._map2write.update(orb.get_map2write())
            self._orbit = orb
            self._orbit.sofb = self

    @property
    def correctors(self):
        """."""
        return self._correctors

    @correctors.setter
    def correctors(self, corrs):
        """."""
        if isinstance(corrs, _BaseCorrectors):
            self._map2write.update(corrs.get_map2write())
            self._correctors = corrs
            self._correctors.sofb = self

    @property
    def matrix(self):
        """."""
        return self._matrix

    @matrix.setter
    def matrix(self, mat):
        """."""
        if isinstance(mat, _BaseMatrix):
            self._map2write.update(mat.get_map2write())
            self._matrix = mat
            self._matrix.sofb = self

    @property
    def havebeam(self):
        """."""
        if self._tests or self.acc != 'SI':
            return True
        return self._havebeam_pv.connected and self._havebeam_pv.value

    def process(self):
        """Run continuously in the main thread."""
        time0 = _time()
        self.status
        tfin = _time()
        dtime = INTERVAL - (tfin-time0)
        if dtime > 0:
            _sleep(dtime)
        else:
            _log.debug('process took {0:f}ms.'.format((tfin-time0)*1000))

    def set_use_pssofb(self, val):
        """Set whether or not closed loop will use PSSOFB.

        Args:
            val (bool): desired state.

        Returns:
            bool: Whether setting was successful.

        """
        if self._thread and self._thread.is_alive():
            msg = 'ERR: Performing some task. Cannot change PSSOFB state!'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        self._loop_use_pssofb = bool(val)
        self.run_callbacks('CorrPSSOFBEnbl-Sts', val)
        return True

    def set_print_every_num_iters(self, value: float) -> bool:
        """Define number of iterations between loop statistics calculation.

        Args:
            value (float): number of iterations to wait.

        Returns:
            bool: whether property was properly set.

        """
        self._loop_print_every_num_iter = int(value)
        self.run_callbacks('LoopPrintEveryNumIters-RB', int(value))
        return True

    def set_fofb_interaction_props(self, prop: str, value: int):
        """Set properties related to FOFB interaction.

        Args:
            prop (str): name of the property
            value (int): value of the property

        Returns:
            bool: Whether property was set.

        """
        value = bool(value)
        if prop.lower().startswith('download'):
            self._download_fofb_kicks = value
            self.run_callbacks('FOFBDownloadKicks-Sts', value)
        elif prop.lower().startswith('update'):
            self._update_fofb_reforb = value
            self.run_callbacks('FOFBUpdateRefOrb-Sts', value)
        elif prop.lower().startswith('null'):
            self._project_onto_fofb_nullspace = value
            self.run_callbacks('FOFBNullSpaceProj-Sts', value)
        elif prop.lower().startswith('zero'):
            self._donot_affect_fofb_bpms = value
            self.run_callbacks('FOFBZeroDistortionAtBPMs-Sts', value)
        else:
            return False
        return True

    def set_fofb_download_perc(self, value: float):
        """Set percentage of kicks to be downloaded from FOFB.

        Args:
            value (float): percentage of kicks. must be in [0, 100].

        Returns:
            bool: Whether property was set.

        """
        value = min(max(value/100, 0), 1)
        self._download_fofb_kicks_perc = value
        self.run_callbacks('FOFBDownloadKicksPerc-RB', value*100)
        return True

    def set_fofb_updatereforb_perc(self, value: float):
        """Set percentage of reference orbit update in FOFB.

        Args:
            value (float): percentage of calculated orbit.
                Must be in [-100, 100].

        Returns:
            bool: Whether property was set.

        """
        value = min(max(value/100, -1), 1)
        self._update_fofb_reforb_perc = value
        self.run_callbacks('FOFBUpdateRefOrbPerc-RB', value*100)
        return True

    def apply_corr(self, code):
        """Apply calculated kicks on the correctors."""
        if self.acc == 'BO':
            msg = 'ERR: Cannot correct orbit for this accelerator.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False

        if self._thread and self._thread.is_alive():
            msg = 'ERR: Loop is Closed or MeasRespMat is On.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        if self._dtheta is None:
            msg = 'ERR: Cannot Apply Kick. Calc Corr first.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        _Thread(
            target=self._apply_corr, kwargs={'code': code},
            daemon=True).start()
        return True

    def calc_correction(self, _):
        """Calculate correction."""
        self.run_callbacks('ApplyDelta-Mon', self._csorb.ApplyDeltaMon.Idle)
        if self._thread and self._thread.is_alive():
            msg = 'ERR: Loop is Closed or MeasRespMat is On.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        _Thread(target=self._calc_correction, daemon=True).start()
        return True

    def set_delta_kick(self, code, dkicks):
        """Calculate correction."""
        if self._thread and self._thread.is_alive():
            msg = 'ERR: Loop is Closed or MeasRespMat is On.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        _Thread(
            target=self._set_delta_kick,
            kwargs={'code': code, 'dkicks': dkicks}, daemon=True).start()
        return True

    def set_respmat_meas_state(self, value):
        """."""
        if value == self._csorb.MeasRespMatCmd.Start:
            self._start_meas_respmat()
        elif value == self._csorb.MeasRespMatCmd.Stop:
            self._stop_meas_respmat()
        elif value == self._csorb.MeasRespMatCmd.Reset:
            self._reset_meas_respmat()
        return True

    def set_auto_corr(self, value):
        """."""
        if self.acc != 'SI':
            msg = 'ERR: Cannot close loop for this accelerator.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False

        if value == self._csorb.LoopState.Closed:
            if self._loop_state == self._csorb.LoopState.Closed:
                msg = 'ERR: Loop is Already closed.'
                self._update_log(msg)
                _log.error(msg[5:])
                return False
            if self._thread and self._thread.is_alive():
                msg = 'ERR: Measuring RespMat or Drive is closed. Stopping!'
                self._update_log(msg)
                _log.error(msg[5:])
                return False
            if not self.havebeam:
                msg = 'ERR: Cannot Correct, We do not have stored beam!'
                self._update_log(msg)
                _log.error(msg[5:])
                return False
            msg = 'Closing the Loop.'
            self._update_log(msg)
            _log.info(msg)
            self._loop_state = value
            self._thread = _Thread(
                target=self._do_auto_corr, daemon=True)
            self._thread.start()
        elif value == self._csorb.LoopState.Open:
            msg = 'Opening the Loop.'
            self._update_log(msg)
            _log.info(msg)
            self._loop_state = value
        return True

    def set_drive_divisor(self, value):
        """."""
        val = abs(int(value))
        self._drive_divisor = min(
            val, self._csorb.MAX_DRIVE_DATA // (3*self._drive_nrcycles))

        self.run_callbacks('DriveFreqDivisor-RB', self._drive_divisor)
        freq = self._csorb.BPMsFreq/self._drive_divisor
        self.run_callbacks('DriveFrequency-Mon', freq)
        self.run_callbacks('DriveDuration-Mon', self._drive_nrcycles/freq)
        return True

    def set_drive_nrcycles(self, value):
        """."""
        val = max(abs(int(value)), 1)
        self._drive_nrcycles = min(
            val, self._csorb.MAX_DRIVE_DATA // (3*self._drive_divisor))

        self.run_callbacks('DriveNrCycles-RB', self._drive_nrcycles)
        freq = self._csorb.BPMsFreq/self._drive_divisor
        self.run_callbacks('DriveDuration-Mon', self._drive_nrcycles/freq)
        return True

    def set_drive_amplitude(self, value):
        """."""
        self._drive_amplitude = value
        self.run_callbacks('DriveAmplitude-RB', value)
        return True

    def set_drive_phase(self, value):
        """."""
        self._drive_phase = value
        self.run_callbacks('DrivePhase-RB', value)
        return True

    def set_drive_corr_index(self, value):
        """."""
        if -self._csorb.nr_corrs < value < self._csorb.nr_corrs:
            self._drive_corr_index = int(value)
            self.run_callbacks('DriveCorrIndex-RB', int(value))
            return True
        return False

    def set_drive_bpm_index(self, value):
        """."""
        if -self._csorb.nr_bpms*2 < value < self._csorb.nr_bpms*2:
            self._drive_bpm_index = int(value)
            self.run_callbacks('DriveBPMIndex-RB', int(value))
            return True
        return False

    def set_drive_type(self, value):
        """."""
        self._drive_type = int(value)
        self.run_callbacks('DriveType-Sts', int(value))
        return True

    def set_drive_state(self, value):
        """."""
        if value == self._csorb.DriveState.Closed:
            if self._drive_state == self._csorb.DriveState.Closed:
                msg = 'ERR: Loop is Already closed.'
                self._update_log(msg)
                _log.error(msg[5:])
                return False
            if self._thread and self._thread.is_alive():
                msg = 'ERR: Measuring RespMat or Loop is Closed. Stopping!'
                self._update_log(msg)
                _log.error(msg[5:])
                return False
            if not self.havebeam:
                msg = 'ERR: Cannot Drive, We do not have stored beam!'
                self._update_log(msg)
                _log.error(msg[5:])
                return False
            msg = 'Closing the Drive Loop.'
            self._update_log(msg)
            _log.info(msg)
            self._drive_state = value
            self._thread = _Thread(
                target=self._do_drive, daemon=True)
            self._thread.start()
        elif value == self._csorb.LoopState.Open:
            msg = 'Opening the Drive Loop.'
            self._update_log(msg)
            _log.info(msg)
            self._drive_state = value
        return True

    def set_auto_corr_frequency(self, value):
        """."""
        bpmfreq = self._csorb.BPMsFreq
        value = bpmfreq / max(int(bpmfreq/value), 1)
        self._loop_freq = value
        self.run_callbacks('LoopFreq-RB', value)
        return True

    def set_pid_gain(self, ctrlr, plane, value):
        """."""
        ctrlr = ctrlr.lower()
        plane = plane.lower()
        self._pid_gains[plane][ctrlr] = float(value)
        self.run_callbacks(
            'LoopPID'+ctrlr.title()+plane.upper()+'-RB', float(value))
        return True

    def set_max_kick(self, plane, value):
        """."""
        self._max_kick[plane] = float(value)
        self.run_callbacks('MaxKick'+plane.upper()+'-RB', float(value))
        return True

    def set_max_delta_kick(self, plane, value):
        """."""
        self._max_delta_kick[plane] = float(value)
        self.run_callbacks('MaxDeltaKick'+plane.upper()+'-RB', float(value))
        return True

    def set_mancorr_gain(self, plane, value):
        """."""
        self._mancorr_gain[plane] = value/100
        msg = 'ManCorrGain{0:s} set to {1:6.2f}'.format(plane.upper(), value)
        self._update_log(msg)
        _log.info(msg)
        self.run_callbacks('ManCorrGain'+plane.upper()+'-RB', value)
        return True

    def set_respmat_kick(self, plane, value):
        """."""
        self._meas_respmat_kick[plane] = value
        self.run_callbacks('MeasRespMatKick'+plane.upper()+'-RB', value)
        return True

    def set_respmat_wait_time(self, value):
        """."""
        self._meas_respmat_wait = value
        self.run_callbacks('MeasRespMatWait-RB', value)
        return True

    def set_max_orbit_dist(self, value):
        """."""
        self._loop_max_orb_distortion = value
        self.run_callbacks('LoopMaxOrbDistortion-RB', value)
        return True

    def _update_status(self):
        self._status = bool(
            self._correctors.status | self._matrix.status | self._orbit.status)
        self.run_callbacks('Status-Mon', self._status)

        if self.acc != 'SI':
            return
        # Update PVs related to interaction with FOFB:
        fofb_state = self.fofb.connected and self.fofb.loop_state
        download = self._download_fofb_kicks and fofb_state
        update = self._update_fofb_reforb and fofb_state
        project = self._project_onto_fofb_nullspace and fofb_state
        donot = self._donot_affect_fofb_bpms and fofb_state
        self.run_callbacks('FOFBDownloadKicks-Mon', download)
        self.run_callbacks('FOFBUpdateRefOrb-Mon', update)
        self.run_callbacks('FOFBNullSpaceProj-Mon', project)
        self.run_callbacks('FOFBZeroDistortionAtBPMs-Mon', donot)

    def _set_delta_kick(self, code, dkicks):
        nr_ch = self._csorb.nr_ch
        nr_chcv = self._csorb.nr_chcv
        self._ref_corr_kicks = self.correctors.get_strength()
        if self._dtheta is None:
            self._dtheta = _np.zeros(self._ref_corr_kicks.size, dtype=float)
        if code == self._csorb.ApplyDelta.CH:
            self._dtheta[:nr_ch] = dkicks
            self.run_callbacks('DeltaKickCH-RB', list(dkicks))
            self.run_callbacks('DeltaKickCH-Mon', list(dkicks))
        elif code == self._csorb.ApplyDelta.CV:
            self._dtheta[nr_ch:nr_chcv] = dkicks
            self.run_callbacks('DeltaKickCV-RB', list(dkicks))
            self.run_callbacks('DeltaKickCV-Mon', list(dkicks))
        elif self.isring and code == self._csorb.ApplyDelta.RF:
            self._dtheta[-1] = dkicks
            self.run_callbacks('DeltaKickRF-RB', float(dkicks))
            self.run_callbacks('DeltaKickRF-Mon', float(dkicks))

    def _apply_corr(self, code):
        self.run_callbacks(
            'ApplyDelta-Mon', self._csorb.ApplyDeltaMon.Applying)
        nr_ch = self._csorb.nr_ch
        if self._dtheta is None:
            msg = 'Err: All kicks are zero.'
            self._update_log(msg)
            _log.warning(msg[6:])
            self.run_callbacks(
                'ApplyDelta-Mon', self._csorb.ApplyDeltaMon.Error)
            return
        dkicks = self._dtheta.copy()
        if code == self._csorb.ApplyDelta.CH:
            dkicks[nr_ch:] = 0
        elif code == self._csorb.ApplyDelta.CV:
            dkicks[:nr_ch] = 0
            if self.isring:
                dkicks[-1] = 0
        elif self.isring and code == self._csorb.ApplyDelta.RF:
            dkicks[:-1] = 0
        msg = f'Applying {self._csorb.ApplyDelta._fields[code]:s} kicks.'
        self._update_log(msg)
        _log.info(msg)

        kicks, dkicks = self._process_kicks(self._ref_corr_kicks, dkicks)
        if kicks is None:
            self.run_callbacks(
                'ApplyDelta-Mon', self._csorb.ApplyDeltaMon.Error)
            return

        if self.acc == 'SI':
            kicks = self._interact_with_fofb_in_apply_kicks(kicks, dkicks)
            if kicks is None:
                self.run_callbacks(
                    'ApplyDelta-Mon', self._csorb.ApplyDeltaMon.Error)
                return

        ret = self.correctors.apply_kicks(kicks)
        if ret is None:
            msg = 'ERR: There is some problem with a corrector!'
            self._update_log(msg)
            _log.error(msg[:5])
            self.run_callbacks(
                'ApplyDelta-Mon', self._csorb.ApplyDeltaMon.Error)
            return
        elif ret == -1:
            msg = 'WARN: Last was not applied yet'
            self._update_log(msg)
            _log.error(msg[:5])
        elif ret == 0:
            msg = 'kicks applied!'
            self._update_log(msg)
            _log.info(msg)
        else:
            msg = f'WARN: {ret:03d} kicks were not applied previously!'
            self._update_log(msg)
            _log.warning(msg[:6])
        self.run_callbacks('ApplyDelta-Mon', self._csorb.ApplyDeltaMon.Done)

    def _stop_meas_respmat(self):
        if not self._measuring_respmat:
            msg = 'ERR: No Measurement ocurring.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        msg = 'Aborting measurement. Wait...'
        self._update_log(msg)
        _log.info(msg)
        self._measuring_respmat = False
        return True

    def _reset_meas_respmat(self):
        if self._measuring_respmat:
            msg = 'Cannot Reset, Measurement in process.'
            self._update_log(msg)
            _log.info(msg)
            return False
        msg = 'Reseting measurement status.'
        self._update_log(msg)
        _log.info(msg)
        self.run_callbacks('MeasRespMat-Mon', self._csorb.MeasRespMatMon.Idle)
        return True

    def _start_meas_respmat(self):
        if self._measuring_respmat:
            msg = 'ERR: Measurement already in process.'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        if self._thread and self._thread.is_alive():
            msg = 'ERR: Loop is Closed or Drive is Running. Stopping!'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        if not self.havebeam:
            msg = 'ERR: Cannot Measure, We do not have stored beam!'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        msg = 'Starting RespMat measurement.'
        self._update_log(msg)
        _log.info(msg)
        self._measuring_respmat = True
        self._thread = _Thread(target=self._do_meas_respmat, daemon=True)
        self._thread.start()
        return True

    def _do_meas_respmat(self):
        self.run_callbacks(
            'MeasRespMat-Mon', self._csorb.MeasRespMatMon.Measuring)
        mat = list()
        orig_kicks = self.correctors.get_strength()
        enbllist = self.matrix.corrs_enbllist
        sum_enbld = sum(enbllist)
        j = 1
        nr_corrs = len(orig_kicks)
        orbzero = _np.zeros(len(self.matrix.bpm_enbllist), dtype=float)
        for i in range(nr_corrs):
            if not self._measuring_respmat:
                self.run_callbacks(
                    'MeasRespMat-Mon', self._csorb.MeasRespMatMon.Aborted)
                msg = 'Measurement stopped.'
                self._update_log(msg)
                _log.info(msg)
                for _ in range(i, nr_corrs):
                    mat.append(orbzero)
                break
            if not self.havebeam:
                self.run_callbacks(
                    'MeasRespMat-Mon', self._csorb.MeasRespMatMon.Aborted)
                msg = 'ERR: Cannot Measure, We do not have stored beam!'
                self._update_log(msg)
                _log.info(msg)
                for _ in range(i, nr_corrs):
                    mat.append(orbzero)
                break
            if not enbllist[i]:
                mat.append(orbzero)
                continue
            msg = '{0:d}/{1:d} -> {2:s}'.format(
                j, sum_enbld, self.correctors.corrs[i].name)
            self._update_log(msg)
            _log.info(msg)
            j += 1

            if i < self._csorb.nr_ch:
                delta = self._meas_respmat_kick['ch']
            elif i < self._csorb.nr_ch + self._csorb.nr_cv:
                delta = self._meas_respmat_kick['cv']
            elif i < self._csorb.nr_corrs:
                delta = self._meas_respmat_kick['rf']

            kicks = _np.array([None, ] * nr_corrs, dtype=float)
            kicks[i] = orig_kicks[i] + delta/2
            self.correctors.apply_kicks(kicks)
            _sleep(self._meas_respmat_wait)
            orbp = self.orbit.get_orbit(reset=True)

            kicks[i] = orig_kicks[i] - delta/2
            self.correctors.apply_kicks(kicks)
            _sleep(self._meas_respmat_wait)
            orbn = self.orbit.get_orbit(reset=True)
            mat.append((orbp-orbn)/delta)

            kicks[i] = orig_kicks[i]
            self.correctors.apply_kicks(kicks)

        mat = _np.array(mat).T
        self.matrix.set_respmat(list(mat.ravel()))
        self.run_callbacks(
            'MeasRespMat-Mon', self._csorb.MeasRespMatMon.Completed)
        self._measuring_respmat = False
        msg = 'RespMat Measurement Completed!'
        self._update_log(msg)
        _log.info(msg)

    def _do_drive(self):
        self.run_callbacks('DriveState-Sts', 1)

        freqdiv = self._drive_divisor
        nrcycles = self._drive_nrcycles
        ampl = self._drive_amplitude
        phase = self._drive_phase / 180 * _np.pi
        corridx = self._drive_corr_index
        bpmidx = self._drive_bpm_index
        dr_type = self._drive_type

        x = _np.arange(freqdiv * nrcycles)
        if dr_type == self._csorb.DriveType.Sine:
            wfm = ampl*_np.sin(2*_np.pi/freqdiv * x + phase)
        elif dr_type == self._csorb.DriveType.Square:
            wfm = _np.zeros(x.size, dtype=float)
            wfm[wfm.size//2:-1] += ampl
        elif dr_type == self._csorb.DriveType.Impulse:
            wfm = _np.zeros(x.size, dtype=float)
            wfm[wfm.size//2] += ampl

        refkicks = self.correctors.get_strength()
        orb = self.orbit.get_orbit(synced=True)
        tim0 = _time()
        data = []
        for idx in range(x.size):
            if self._drive_state != self._csorb.DriveState.Closed:
                break
            if not self.havebeam:
                msg = 'ERR: Cannot Drive, We do not have stored beam!'
                self._update_log(msg)
                _log.info(msg)
                break

            orb = self.orbit.get_orbit(synced=True)
            kicks = refkicks.copy()
            tim = _time() - tim0
            kicks[corridx] += wfm[idx]
            ret = self.correctors.apply_kicks(kicks)
            data.extend([tim, kicks[corridx], orb[bpmidx]])

            if ret == -2:
                self._drive_state = self._csorb.DriveState.Open
                self.run_callbacks(
                    'DriveState-Sel', self._csorb.DriveState.Open)
                break
            elif ret == -1:
                # means that correctors are not ready yet
                # skip this iteration
                continue
        else:
            self._drive_state = self._csorb.DriveState.Open
            self.run_callbacks(
                'DriveState-Sel', self._csorb.DriveState.Open)

        ret = self.correctors.apply_kicks(refkicks)
        if len(data) < 6:
            data.extend((6-len(data)) * [0.0])
        self.run_callbacks('DriveData-Mon', data)
        msg = 'Drive Loop opened!'
        self._update_log(msg)
        _log.info(msg)
        self.run_callbacks('DriveState-Sts', 0)

    def _do_auto_corr(self):
        self.run_callbacks('LoopState-Sts', 1)
        times, rets = [], []
        tim0 = _time()
        bpmsfreq = self._csorb.BPMsFreq
        zer = _np.zeros(self._csorb.nr_corrs, dtype=float)
        self._pid_errs = [zer, zer.copy(), zer.copy()]
        fofb = self.fofb
        refx0 = refy0 = None
        if fofb.connected:
            refx0 = fofb.refx
            refy0 = fofb.refy

        if self._loop_use_pssofb:
            msg = 'Turning on PSSOFB...'
            self._update_log(msg)
            _log.info(msg)
            self.correctors.use_pssofb = True
            msg = 'PSSOFB ready!'
            self._update_log(msg)
            _log.info(msg)

            if self.correctors.sync_kicks != self._csorb.CorrSync.Off:
                msg = 'Setting Trigger to listen to EVG Clock...'
                self._update_log(msg)
                _log.info(msg)
                self.run_callbacks('CorrSync-Sts', self._csorb.CorrSync.Clock)
                self.correctors.set_corrs_mode(self._csorb.CorrSync.Clock)
                msg = 'Trigger ready!'
                self._update_log(msg)
                _log.info(msg)

        tims = []
        while self._loop_state == self._csorb.LoopState.Closed:
            if not self.havebeam:
                msg = 'ERR: Cannot Correct, We do not have stored beam!'
                self._update_log(msg)
                _log.info(msg)
                break
            itern = len(times)
            self.run_callbacks('LoopNumIters-Mon', itern)
            if itern >= self._loop_print_every_num_iter:
                _Thread(
                    target=self._print_auto_corr_info,
                    args=(times, rets, _time()-tim0), daemon=True).start()
                times, rets = [], []
                tim0 = _time()

            interval = 1/self._loop_freq
            use_pssofb = self.correctors.use_pssofb
            norbs = 1
            if use_pssofb:
                norbs = max(int(bpmsfreq*interval), 1)
            elif tims:
                # if not pssofb wait for interval to be satisfied
                dtime = tims[0] - tims[-1]
                dtime += interval
                _sleep(max(dtime, 0))

            tims = []
            tims.append(_time())
            orb = self.orbit.get_orbit(synced=True)
            for i in range(1, norbs):
                interval = 1/self._loop_freq
                norbs_up = max(int(bpmsfreq*interval), 1)
                if i >= norbs_up:
                    break
                orb = self.orbit.get_orbit(synced=True)

            if self._tests:
                orb *= 0
                orb += _np.random.rand(orb.size)
                orb -= orb.mean()  # avoid RF integration error.
                orb *= 2 * 3  # Maximum orbit distortion of 3 um

            tims.append(_time())

            self._ref_corr_kicks = self.correctors.get_strength()
            tims.append(_time())

            orb = self._interact_with_fofb_in_calc_kicks(orb)

            dkicks = self.matrix.calc_kicks(orb)
            tims.append(_time())

            if not self._check_valid_orbit(orb):
                self._loop_state = self._csorb.LoopState.Open
                self.run_callbacks('LoopState-Sel', 0)
                break

            dkicks = self._process_pid(dkicks, interval)

            kicks, dkicks = self._process_kicks(
                self._ref_corr_kicks, dkicks, apply_gain=False)
            if kicks is None:
                self._loop_state = self._csorb.LoopState.Open
                self.run_callbacks('LoopState-Sel', 0)
                break

            kicks = self._interact_with_fofb_in_apply_kicks(
                kicks, dkicks, refx0, refy0)
            if kicks is None:
                self._loop_state = self._csorb.LoopState.Open
                self.run_callbacks('LoopState-Sel', 0)
                break
            tims.append(_time())

            ret = self.correctors.apply_kicks(kicks)
            rets.append(ret)
            tims.append(_time())
            tims.append(tims[1])  # to compute total time - get_orbit
            times.append(tims)
            # if ret == -2:
            if ret < 0:  # change here for debug
                self._loop_state = self._csorb.LoopState.Open
                self.run_callbacks('LoopState-Sel', 0)
                break
            elif ret == -1:
                # means that correctors are not ready yet
                # skip this iteration
                continue

        if self._loop_use_pssofb:
            msg = 'Turning off PSSOFB...'
            self._update_log(msg)
            _log.info(msg)
            self.correctors.use_pssofb = False
            msg = 'PSSOFB is off!'
            self._update_log(msg)
            _log.info(msg)

            if self.correctors.sync_kicks != self._csorb.CorrSync.Off:
                msg = 'Setting Trigger to listen to Event...'
                self._update_log(msg)
                _log.info(msg)
                self.run_callbacks('CorrSync-Sts', self._csorb.CorrSync.Event)
                self.correctors.set_corrs_mode(self._csorb.CorrSync.Event)
                msg = 'Trigger ready!'
                self._update_log(msg)
                _log.info(msg)

        msg = 'Loop opened!'
        self._update_log(msg)
        _log.info(msg)
        self.run_callbacks('LoopState-Sts', 0)

    def _process_pid(self, dkicks, interval):
        """Velocity algorithm of PID."""
        if dkicks is None:
            return None

        self._pid_errs.append(dkicks.copy())
        del self._pid_errs[0]

        gains = self._pid_gains
        errs = self._pid_errs
        nr_ch = self._csorb.nr_ch
        slcs = {'ch': slice(None, nr_ch), 'cv': slice(nr_ch, None)}
        if self.isring:
            slcs = {
                'ch': slice(None, nr_ch),
                'cv': slice(nr_ch, -1),
                'rf': slice(-1, None)}
        for pln in sorted(slcs.keys()):
            slc = slcs[pln]
            gin = gains[pln]
            kpt = gin['kp']
            kdt = gin['kd']/interval
            kit = gin['ki']*interval

            if self._tests:
                # Do not use integrators when testing without beam:
                kit = 0
                if pln == 'rf':
                    # Do not use RF when testing without beam:
                    kpt = kdt = 0

            qq0 = kpt + kdt + kit
            qq1 = -kpt - 2*kdt
            qq2 = kdt
            dkicks[slc] *= qq0
            dkicks[slc] += qq1*errs[-2][slc]  # previous error
            dkicks[slc] += qq2*errs[-3][slc]  # pre-previous error
        return dkicks

    def _print_auto_corr_info(self, times, rets, dtim):
        """."""
        self.run_callbacks('LoopEffectiveRate-Mon', len(times)/dtim)

        rets = _np.array(rets)
        ok_ = _np.sum(rets == 0) / rets.size * 100
        tout = _np.sum(rets == -1) / rets.size * 100
        bo_diff = rets > 0
        diff = _np.sum(bo_diff) / rets.size * 100
        _log.info('PERFORMANCE:')
        self.run_callbacks('LoopPerfItersOk-Mon', ok_)
        self.run_callbacks('LoopPerfItersTOut-Mon', tout)
        self.run_callbacks('LoopPerfItersDiff-Mon', diff)
        psmax = psavg = psstd = 0
        if diff:
            drets = rets[bo_diff]
            psmax = drets.max()
            psavg = drets.mean()
            psstd = drets.std()
        self.run_callbacks('LoopPerfDiffNrPSMax-Mon', psmax)
        self.run_callbacks('LoopPerfDiffNrPSAvg-Mon', psavg)
        self.run_callbacks('LoopPerfDiffNrPSStd-Mon', psstd)

        dtimes = _np.diff(times, axis=1).T * 1000
        dtimes[-1] *= -1
        max_ = dtimes.max(axis=1)
        min_ = dtimes.min(axis=1)
        avg_ = dtimes.mean(axis=1)
        std_ = dtimes.std(axis=1)
        labs = ['GetO', 'GetK', 'Calc', 'Proc', 'App', 'Tot']
        for i, lab in enumerate(labs):
            self.run_callbacks(f'LoopPerfTim{lab:s}Max-Mon', max_[i])
            self.run_callbacks(f'LoopPerfTim{lab:s}Min-Mon', min_[i])
            self.run_callbacks(f'LoopPerfTim{lab:s}Avg-Mon', avg_[i])
            self.run_callbacks(f'LoopPerfTim{lab:s}Std-Mon', std_[i])

    def _check_valid_orbit(self, orbit):
        conn = _np.array([bpm.connected for bpm in self._orbit.bpms])
        conn = _np.tile(conn, [2, ])
        enbl = self._matrix.bpm_enbllist
        nconn = ~conn & enbl
        if _np.any(nconn):
            names = self._csorb.bpm_names
            leng = len(names)
            for i, nco in enumerate(nconn):
                if nco:
                    msg = f'ERR: {names[i%leng]} not connected!'
                    self._update_log(msg)
                    _log.error(msg[5:])
            return False

        maxo = _np.abs(orbit[enbl]).max()
        if maxo > self._loop_max_orb_distortion:
            msg = 'ERR: Orbit distortion above threshold!'
            self._update_log(msg)
            _log.error(msg[5:])
            return False
        return True

    def _interact_with_fofb_in_calc_kicks(self, orb):
        fofb = self.fofb

        if self._donot_affect_fofb_bpms and fofb.loop_state:
            enbllist = _np.r_[fofb.bpmxenbl, fofb.bpmyenbl]
            orb[enbllist] = 0.0

        if self._project_onto_fofb_nullspace and fofb.loop_state:
            # this approach is similar to what is proposed by APS:
            # https://www.aps.anl.gov/sites/www.aps.anl.gov/files/APS-Uploads/Workshops/BES-Light-Sources/Nick%20Sereno%20-%20Fast%20Orbit%20Feedback%20at%20APS.pdf
            imat_fofb = fofb.invrespmat_mon
            imat_fofb[-1] *= 0  # RF correction is never applied by FOFB.
            orb -= _np.dot(fofb.respmat, _np.dot(imat_fofb, orb))
        return orb

    def _interact_with_fofb_in_apply_kicks(
            self, kicks, dkicks, refx=None, refy=None):
        fofb = self.fofb

        # if refx is None or refy is None:
        #     refx = fofb.refx
        #     refy = fofb.refy

        if self._update_fofb_reforb and fofb.loop_state:
            dorb = self.matrix.estimate_orbit_variation(dkicks)
            dorb *= self._update_fofb_reforb_perc
            # NOTE: According to my understanding of SOLEIL's paper on this
            # subject:
            # https://accelconf.web.cern.ch/d09/papers/mooc01.pdf
            # https://accelconf.web.cern.ch/d09/talks/mooc01_talk.pdf
            # this is what they do there:
            fofb.refx -= dorb[:dorb.size//2]
            fofb.refy -= dorb[dorb.size//2:]
            # But it seems this may also be a possibility:
            # fofb.refx = refx - dorb[:dorb.size//2]
            # fofb.refy = refy - dorb[dorb.size//2:]
            fofb.cmd_fofbctrl_syncreforb()

        if self._download_fofb_kicks and fofb.loop_state:
            # NOTE: Do not download kicks from correctors not in the loop:
            kickch = fofb.kickch.copy()
            kickcv = fofb.kickcv.copy()
            kickch[~fofb.chenbl] = 0
            kickcv[~fofb.cvenbl] = 0

            kicks_fofb = _np.r_[kickch, kickcv, 0]
            dorb = _np.dot(fofb.respmat, kicks_fofb)
            # NOTE: calc_kicks return the kicks to correct dorb, which means
            # that a minus sign is already applied by this method. To negate
            # this correction, we need an extra minus sign here:
            dkicks2 = self.matrix.calc_kicks(dorb)
            dkicks2 *= -self._download_fofb_kicks_perc

            kicks, dkicks2 = self._process_kicks(
                self._ref_corr_kicks, dkicks + dkicks2, apply_gain=False)
        return kicks

    def _calc_correction(self):
        msg = 'Getting the orbit.'
        self._update_log(msg)
        _log.info(msg)
        orb = self.orbit.get_orbit()

        if self.acc == 'SI':
            orb = self._interact_with_fofb_in_calc_kicks(orb)

        msg = 'Calculating kicks.'
        self._update_log(msg)
        _log.info(msg)
        self._ref_corr_kicks = self.correctors.get_strength()
        dkicks = self.matrix.calc_kicks(orb)
        if dkicks is not None:
            self._dtheta = dkicks
        msg = 'Kicks calculated!'
        self._update_log(msg)
        _log.info(msg)

    def _process_kicks(self, kicks, dkicks, apply_gain=True):
        if dkicks is None:
            return None, dkicks

        # keep track of which dkicks were originally different from zero:
        newkicks = _np.full(dkicks.shape, _np.nan, dtype=float)
        apply_idcs = ~_compare_kicks(dkicks, 0)
        if not apply_idcs.any():
            return newkicks, dkicks

        nr_ch = self._csorb.nr_ch
        slcs = {'ch': slice(None, nr_ch), 'cv': slice(nr_ch, None)}
        if self.isring:
            slcs = {
                'ch': slice(None, nr_ch),
                'cv': slice(nr_ch, -1),
                'rf': slice(-1, None)}
        for pln in sorted(slcs.keys()):
            slc = slcs[pln]
            idcs_pln = apply_idcs[slc]
            if not idcs_pln.any():
                continue
            dk_slc = dkicks[slc][idcs_pln]
            k_slc = kicks[slc][idcs_pln]
            fac1 = self._mancorr_gain[pln] if apply_gain else 1
            dk_slc *= fac1

            # Check if any kick is larger than the maximum allowed:
            ind = (_np.abs(k_slc) >= self._max_kick[pln]).nonzero()[0]
            if ind.size:
                msg = 'ERR: Kicks above MaxKick{0:s}.'.format(pln.upper())
                self._update_log(msg)
                _log.error(msg[5:])
                return None, dkicks

            # Check if any delta kick is larger the maximum allowed
            max_delta_kick = _np.max(_np.abs(dk_slc))
            fac2 = 1.0
            if max_delta_kick > self._max_delta_kick[pln]:
                fac2 = self._max_delta_kick[pln]/max_delta_kick
                dk_slc *= fac2
                percent = fac1 * fac2 * 100
                msg = 'WARN: reach MaxDeltaKick{0:s}. Using {1:5.2f}%'.format(
                    pln.upper(), percent)
                self._update_log(msg)
                _log.warning(msg[6:])

            # Check if any kick + delta kick is larger than the maximum allowed
            max_kick = _np.max(_np.abs(k_slc + dk_slc))
            fac3 = 1.0
            if max_kick > self._max_kick[pln]:
                que = _np.ones((2, k_slc.size), dtype=float)
                # perform the modulus inequality:
                que[0, :] = (-self._max_kick[pln] - k_slc) / dk_slc
                que[1, :] = (self._max_kick[pln] - k_slc) / dk_slc
                # since we know that any initial kick is lesser than max_kick
                # from the previous comparison, at this point each column of Q
                # has a positive and a negative value. We must consider only
                # the positive one and take the minimum value along the columns
                # to be the multiplicative factor:
                que = _np.max(que, axis=0)
                fac3 = min(_np.min(que), 1.0)
                if fac3 < 1e-4:
                    msg = f'ERR: Some {pln.upper():s} Corr is saturated.'
                    self._update_log(msg)
                    _log.error(msg[5:])
                    return None, dkicks
                dk_slc *= fac3
                percent = fac1 * fac2 * fac3 * 100
                msg = 'WARN: reach MaxKick{0:s}. Using {1:5.2f}%'.format(
                    pln.upper(), percent)
                self._update_log(msg)
                _log.warning(msg[6:])

            dkicks[slc][idcs_pln] = dk_slc

        newkicks[apply_idcs] = kicks[apply_idcs] + dkicks[apply_idcs]
        return newkicks, dkicks



"""Siriuspy unittest package."""

# make sure tests are run without access to sirius servers!
import os as _os
from . import mock_servweb

_os.environ['SIRIUS_URL_RBAC_AUTH'] = ''
_os.environ['SIRIUS_URL_RBAC'] = ''
_os.environ['SIRIUS_URL_NS'] = ''
_os.environ['SIRIUS_URL_CCDB'] = ''
_os.environ['SIRIUS_URL_CABLES'] = ''
_os.environ['SIRIUS_URL_CONSTS'] = ''
_os.environ['SIRIUS_URL_LOGBOOK'] = ''
_os.environ['SIRIUS_URL_CONFIGDB'] = ''



"""."""

from functools import partial as _part
from threading import Event as _Event
import numpy as _np
from epics import PV as _PV

import mathphys.constants as _consts

from ...magnet.factory import NormalizerFactory as _NormFact

from ..util import BaseClass as _BaseClass, ProcessImage as _ProcessImage

C = _consts.light_speed
E0 = _consts.electron_rest_energy / _consts.elementary_charge * 1e-6  # in MeV


class CalcEmmitance(_BaseClass):
    """."""

    X = 0
    Y = 1
    PLACES = ('li', 'tb-qd2a', 'tb-qf2a')

    def __init__(self):
        """."""
        super().__init__()
        self._place = 'LI'
        self._quadname = ''
        self._quadlen = 0.0
        self._distance = 0.0
        self._select_experimental_setup()
        self._beamsize = _np.array([], dtype=float)
        self._beamsize_fit = _np.array([], dtype=float)
        self._currents = _np.array([], dtype=float)
        self._quadstren = _np.array([], dtype=float)
        self._plane = self.Plane.X
        self._energy = 0.15
        self._emittance = 0.0
        self._beta = 0.0
        self._alpha = 0.0
        self._gamma = 0.0

    def get_map2write(self):
        """."""
        return {
            'Place-Sel': _part(self.write, 'place'),
            'Plane-Sel': _part(self.write, 'plane'),
            }

    def get_map2read(self):
        """."""
        return {
            'Place-Sts': _part(self.read, 'place'),
            'Plane-Sts': _part(self.read, 'plane'),
            'QuadName-Mon': _part(self.read, 'quadname'),
            'QuadLen-Mon': _part(self.read, 'quadlen'),
            'Distance-Mon': _part(self.read, 'distance'),
            'BeamSizes-Mon': _part(self.read, 'beamsize'),
            'BeamSizesFit-Mon': _part(self.read, 'beamsize_fit'),
            'QuadCurrents-Mon': _part(self.read, 'currents'),
            'QuadStrens-Mon': _part(self.read, 'quadstren'),
            'Beta-Mon': _part(self.read, 'beta'),
            'Alpha-Mon': _part(self.read, 'alpha'),
            'Gamma-Mon': _part(self.read, 'gamma'),
            'Emittance-Mon': _part(self.read, 'emittance'),
            'NormEmittance-Mon': _part(self.read, 'norm_emittance'),
            }

    @property
    def place(self):
        """."""
        return self._place

    @place.setter
    def place(self, val):
        """."""
        if isinstance(val, str) and val.lower() in self.PLACES:
            self._place = val
            self._select_experimental_setup()
            self._perform_analysis()

    @property
    def plane(self):
        """."""
        return self._plane

    @plane.setter
    def plane(self, val):
        """."""
        if int(val) in self.Plane:
            self._plane = int(val)
        self._perform_analysis()

    @property
    def plane_str(self):
        """."""
        return self.Plane._fields[self._plane]

    @plane_str.setter
    def plane_str(self, val):
        """."""
        if val in self.Plane._fields:
            self.plane = self.Plane._fields.index(val)

    @property
    def quadname(self):
        """."""
        return self._quadname

    @property
    def quadlen(self):
        """."""
        return self._quadlen

    @property
    def distance(self):
        """."""
        return self._distance

    @property
    def beamsize(self):
        """."""
        return self._beamsize.copy()

    @property
    def beamsize_fit(self):
        """."""
        return self._beamsize_fit.copy()

    @property
    def currents(self):
        """."""
        return self._currents.copy()

    @property
    def quadstren(self):
        """."""
        return self._quadstren.copy()

    @property
    def beta(self):
        """."""
        return self._beta

    @property
    def alpha(self):
        """."""
        return self._alpha

    @property
    def gamma(self):
        """."""
        return self._gamma

    @property
    def emittance(self):
        """."""
        return self._emittance * 1e6  # in mm.mrad

    @property
    def norm_emittance(self):
        """."""
        return self.emittance * self._energy / E0

    def set_data(self, beam_sizes, currents):
        """."""
        if None in {currents, beam_sizes}:
            return False
        if isinstance(beam_sizes, (int, float)):
            beam_sizes = [beam_sizes]
        if isinstance(currents, (int, float)):
            currents = [currents]
        beam_sizes = _np.array(beam_sizes, dtype=float)
        currents = _np.array(currents, dtype=float)
        if beam_sizes.size != currents.size:
            return False
        self._beamsize = beam_sizes
        self._currents = currents
        self._perform_analysis()
        return True

    def _select_experimental_setup(self):
        """."""
        if self.place.lower().startswith('li'):
            self._quadname = 'LI-01:PS-QF3'
            self._quadlen = 0.112
            self._distance = 2.8775
        if self.place.lower().startswith('tb-qd2a'):
            self._quadname = 'TB-02:PS-QD2A'
            self._quadlen = 0.1
            self._distance = 6.904
        if self.place.lower().startswith('tb-qf2a'):
            self._quadname = 'TB-02:PS-QF2A'
            self._quadlen = 0.1
            self._distance = 6.534
        self._conv2kl = _NormFact.create(self._quadname)

    def _perform_analysis(self):
        """."""
        if not self._beamsize.size:
            return
        self._update_quad_strength()
        self._trans_matrix_analysis()
        self._thin_lens_approx()

    def _update_quad_strength(self):
        """."""
        KL = self._conv2kl.conv_current_2_strength(
            self._currents, strengths_dipole=self._energy)
        self._quadstren = KL/self._quadlen
        if self._plane == self.Plane.Y:
            self._quadstren *= -1

    def _thin_lens_approx(self):
        """."""
        fit = _np.polyfit(self._quadstren, self._beamsize**2, 2)
        self._beamsize_fit = _np.sqrt(_np.polyval(fit, self._quadstren))
        # d = self.distance + self.quad_length/2
        # l = self.quad_length
        # s_11 = fit[0]/(d*l)**2
        # s_12 = (-fit[1]-2*d*l*s_11)/(2*l*d*d)
        # s_22 = (fit[2]-s_11-2*d*s_12)/d**2
        # self._update_twiss(s_11, s_12, s_22)

    def _trans_matrix_analysis(self):
        """."""
        R = self._get_trans_mat()
        pseudo_inv = (_np.linalg.inv(_np.transpose(R) @ R) @ _np.transpose(R))
        [s_11, s_12, s_22] = pseudo_inv @ (self._beamsize**2)
        # s_11, s_12, s_22 = np.linalg.lstsq(R, self._beamsize**2)[0]
        self._update_twiss(s_11, s_12, s_22)

    def _get_trans_mat(self):
        """."""
        R = _np.zeros((self._quadstren.size, 2, 2))
        Rd = self.gettransmat('drift', L=self._distance)
        for i, k1 in enumerate(self._quadstren):
            Rq = self.gettransmat('quad', L=self._quadlen, K1=k1)
            R[i] = _np.dot(Rd, Rq)
        R11 = R[:, 0, 0].reshape(-1)
        R12 = R[:, 0, 1].reshape(-1)
        R = _np.column_stack((R11*R11, 2*R11*R12, R12*R12))
        return R

    def _update_twiss(self, s_11, s_12, s_22):
        """."""
        self._emittance = _np.sqrt(abs(s_11*s_22 - s_12*s_12))
        self._beta = s_11 / self._emittance
        self._alpha = -s_12 / self._emittance
        self._gamma = s_22 / self._emittance

    @staticmethod
    def gettransmat(elem, L, K1=None, B=None):
        """."""
        matrix = _np.eye(2)
        if elem.lower().startswith('qu') and K1 is not None and K1 == 0:
            elem = 'drift'
        if elem.lower().startswith('dr'):
            matrix = _np.array([[1, L], [0, 1], ])
        elif elem.lower().startswith('qu') and K1 is not None:
            vkq = _np.sqrt(abs(K1))
            if K1 > 0:
                cos = _np.cos(vkq*L)
                sin = _np.sin(vkq*L)
                m11, m12, m21 = cos, 1/vkq*sin, -vkq*sin
            else:
                hcos = _np.cosh(vkq*L)
                hsin = _np.sinh(vkq*L)
                m11, m12, m21 = hcos, 1/vkq*hsin, vkq*hsin
            matrix = _np.array([[m11, m12], [m21, m11], ])
        return matrix


class MeasEmmitance(_BaseClass):
    """."""

    X = 0
    Y = 1
    PLACES = ('li', 'tb-qd2a', 'tb-qf2a')

    def __init__(self):
        """."""
        super().__init__()
        self._measuring = _Event()
        self.emittance_calculator = CalcEmmitance()
        self.image_processor = _ProcessImage()
        self.image_processor.readingorder = \
            self.image_processor.ReadingOrder.CLike
        self._place = 'li'
        self._select_experimental_setup()

    def get_map2write(self):
        """."""
        database = dict()
        dic_ = self.image_processor.get_map2write()
        dic_.update(self.emittance_calculator.get_map2write())
        dic_.update({'MeasureCtrl-Sel': _part(self.write, 'measuring')})
        return {k: v for k, v in dic_.items() if k in database}

    def get_map2read(self):
        """."""
        database = dict()
        dic_ = self.image_processor.get_map2read()
        dic_.update(self.emittance_calculator.get_map2read())
        dic_.update({'MeasureCtrl-Sts': _part(self.read, 'measuring')})
        return {k: v for k, v in dic_.items() if k in database}

    @property
    def place(self):
        """."""
        return self._place

    @place.setter
    def place(self, val):
        """."""
        if val in self.PLACES:
            self._place = val
            self._select_experimental_setup()

    def _select_experimental_setup(self):
        """."""
        self.emittance_calculator.place = self._place
        if self._place.lower().startswith('li'):
            prof = 'LA-BI:PRF5'
            self._image_source = _PV(prof+':RAW:ArrayData')
            self._width_source = _PV(
                prof+':ROI:MaxSizeX_RBV', callback=self._update_width)
            self._coefx = _PV(
                prof+':X:Gauss:Coef',
                callback=self._update_coefx)
            self._coefy = _PV(
                prof+':Y:Gauss:Coef',
                callback=self._update_coefy)
            self.quad_I_sp = _PV('LI-01:PS-QF3:Current-SP')
            self.quad_I_rb = _PV('LI-01:PS-QF3:Current-Mon')
        elif self._place.lower().startswith('tb'):
            self._image_source = _PV('TB-02:DI-Scrn-2:ImgData-Mon')
            self._width_source = _PV(
                'TB-02:DI-Scrn-2:ImgMaxWidth-Cte', callback=self._update_width)
            self._coefx = _PV(
                'TB-02:DI-ScrnCam-2:ImgScaleFactorX-RB',
                callback=self._update_coefx)
            self._coefy = _PV(
                'TB-02:DI-ScrnCam-2:ImgScaleFactorY-RB',
                callback=self._update_coefy)
            quad = self.emittance_calculator.quadname
            self.quad_I_sp = _PV(quad + ':Current-SP')
            self.quad_I_rb = _PV(quad + ':Current-RB')

    def _update_coefx(self, pvname, value, **kwargs):
        """."""
        if value is None:
            return
        self.image_processor.pxl2mmscalex = value

    def _update_coefy(self, pvname, value, **kwargs):
        """."""
        if value is None:
            return
        self.image_processor.pxl2mmscaley = value

    def _update_width(self, pvname, value, **kwargs):
        """."""
        if isinstance(value, (float, int)):
            self.image_processor.imagewidth = int(value)

    def _acquire_data(self):
        """."""
        samples = self.spbox_samples.value()
        nsteps = self.spbox_steps.value()
        I_ini = self.spbox_I_ini.value()
        I_end = self.spbox_I_end.value()

        pl = 'y' if self.cbbox_plane.currentIndex() else 'x'
        curr_list = _np.linspace(I_ini, I_end, nsteps)
        sigma = []
        I_meas = []
        for i, I in enumerate(curr_list):
            print('setting Quadrupole to ', I)
            self.quad_I_sp.put(I, wait=True)
            self._measuring.wait(5 if i else 15)
            j = 0
            I_tmp = []
            sig_tmp = []
            while j < samples:
                if self._measuring.is_set():
                    print('Stopped')
                    return
                print('measuring sample', j)
                I_now = self.quad_I_rb.value
                cen_x, sigma_x, cen_y, sigma_y = self.plt_image.get_params()
                mu, sig = (cen_x, sigma_x) if pl == 'x' else (cen_y, sigma_y)
                max_size = self.spbox_threshold.value()*1e-3
                if sig > max_size:
                    self._measuring.wait(1)
                    continue
                I_tmp.append(I_now)
                sig_tmp.append(abs(sig))
                self._measuring.wait(0.5)
                j += 1
            ind = np.argsort(sig_tmp)
            I_tmp = np.array(I_tmp)[ind]
            sig_tmp = np.array(sig_tmp)[ind]
            I_meas.extend(I_tmp[6:-6])
            sigma.extend(sig_tmp[6:-6])
        self._measuring.set()
        print('Finished!')
        self.I_meas = I_meas
        self.sigma = sigma
        self.plane_meas = pl



"""Excitation Data."""

import numpy as _np

from .. import clientweb as _web

from . import util as _util


class ExcitationData:
    """ExcitationData Class.

    This class implements access to excitation data paremeters and
    current - integrated field conversions.
    """

    def __init__(self, filename_web=None, filename=None, text=None):
        """Init method."""
        # define attributes
        self.label = None
        self._harmonics = None
        self._main_multipole_harmonic = None
        self._main_multipole_type = None
        self.rescaling_factor = None
        self.column_units = None
        self.currents = None
        self.multipoles = None

        # init attributes
        self._init()

        # check arguments
        args = (filename, text, filename_web)
        if sum((arg is not None for arg in args)) != 1:
            raise Exception('ambiguous constructor arguments')

        if text:
            self._read_text(text)
        elif filename:
            self._read_file(filename)
        else:
            self._read_webs(filename_web)

    def _init(self):

        self.label = ''
        self._harmonics = list()
        self._main_multipole_harmonic = None
        self._main_multipole_type = None
        self.rescaling_factor = 1.0
        self.column_units = ''
        self.currents = list()
        self.multipoles = dict()

    # --- properties ---

    @property
    def harmonics(self):
        """Return list with multipole harmonics."""
        return [harm for harm in self._harmonics]

    @property
    def main_multipole_harmonic(self):
        """Return main multipole."""
        return self._main_multipole_harmonic

    @property
    def main_multipole_type(self):
        """Return string indicating type of main multipoel."""
        return self._main_multipole_type

    @property
    def current_limits(self):
        """Return current limits."""
        return self._current_limits

    # --- public methods ---

    def check_valid_curr(self, value):
        """Check if current is within limits of excitation data."""
        return self.currents[0] <= value <= self.currents[-1]

    def check_valid_mult(self, value, harmonic, multipole_type):
        """Check if multipole is within limits of excitation data."""
        multipoles = self.multipoles[multipole_type][harmonic]
        return min(multipoles) <= value <= max(multipoles)

    def interp_curr2mult(self, currents, only_main_harmonic=False):
        """Interpolate multipoles for current values."""
        curr = self.currents
        multipoles = {'normal': {}, 'skew': {}}
        harmonics = (self.main_multipole_harmonic, ) if \
            only_main_harmonic else self.harmonics
        if _np.isscalar(currents):
            currents = _np.array([currents])
            for harm in harmonics:
                # normal component
                mpole = self.multipoles['normal'][harm]
                interp = _util.linear_interpolation(currents, curr, mpole)
                multipoles['normal'][harm] = interp[0]
                # skew component
                mpole = self.multipoles['skew'][harm]
                interp = _util.linear_interpolation(currents, curr, mpole)
                multipoles['skew'][harm] = interp[0]
        else:
            currents = _np.array(currents)
            for harm in harmonics:
                # normal component
                mpole = self.multipoles['normal'][harm]
                interp = _util.linear_interpolation(currents, curr, mpole)
                multipoles['normal'][harm] = interp
                # skew component
                mpole = self.multipoles['skew'][harm]
                interp = _util.linear_interpolation(currents, curr, mpole)
                multipoles['skew'][harm] = interp
        return multipoles

    def interp_mult2curr(self, multipoles, harmonic, multipole_type):
        """Interpolate current from a specific multipole value."""
        # sort correctly tabulated lists
        mpole = self.multipoles[multipole_type][harmonic]
        curr = self.currents
        if mpole[-1] <= mpole[0]:
            mpole, curr = mpole[::-1], self.currents[::-1]

        # do conversion
        if _np.isscalar(multipoles):
            multipoles = _np.array([multipoles])
            interp = _util.linear_interpolation(multipoles, mpole, curr)
            currents = interp[0]
        else:
            multipoles = _np.array(multipoles)
            interp = _util.linear_interpolation(multipoles, mpole, curr)
            currents = interp
        return currents

    # --- special methods ---

    def __str__(self):
        """Str method."""
        str_h = [str(harm) for harm in self._harmonics]
        rst = ''
        rst += '# Magnet Excitation Data file' + '\n'
        rst += '# ===========================' + '\n'
        rst += '#' + '\n'
        rst += '# label                      ' + self.label + '\n'
        rst += '# harmonics                  ' + ' '.join(str_h) + '\n'
        rst += '# main_multipole_harmonic    ' + \
            str(self._main_multipole_harmonic) + '\n'
        rst += '# main_multipole_type        ' + \
            self._main_multipole_type + '\n'
        rst += '# column_units               ' + 'A  '
        for harm in self._harmonics:
            rst += '{0:s} {1:s}  '.format(
                _util.get_multipole_si_units(harm),
                _util.get_multipole_si_units(harm))
        rst += '\n'
        rst += '\n'
        for i in range(len(self.currents)):
            rst += '{0:>+9.3f}  '.format(self.currents[i])
            for harm in self._harmonics:
                rst += '{0:+.5e} {1:+.5e}  '.format(
                    self.multipoles['normal'][harm][i],
                    self.multipoles['skew'][harm][i])
            rst += '\n'
        return rst

    # --- private methods ---

    def _process_comment_line(self, line):
        if line[1:].strip():
            token, *words = line[1:].split()
            if token.lower() == 'label':
                self.label = words[0]
            elif token.lower() == 'harmonics':
                self._harmonics = [int(vec) for vec in words]
            elif token.lower() == 'main_harmonic':
                self._main_multipole_harmonic = int(words[0])
                self._main_multipole_type = words[1].lower()
            elif token.lower() == 'main_multipole_harmonic':
                self._main_multipole_harmonic = int(words[0])
            elif token.lower() == 'main_multipole_type':
                self._main_multipole_type = words[0]
            elif token.lower() == 'units':
                self.column_units = ' '.join(words)
            elif token.lower() == 'column_units':
                self.column_units = ' '.join(words)
            elif token.lower() == 'rescaling_factor':
                self.rescaling_factor = float(words[0])

    def _process_data(self, line):
        if not self.multipoles:
            self.multipoles['normal'] = \
                {harm: [] for harm in self._harmonics}
            self.multipoles['skew'] = {harm: [] for harm in self._harmonics}
        cur, *exc = line.split()
        self.currents.append(float(cur))
        for j in range(len(self._harmonics)):
            harm = self._harmonics[j]
            nmpole = self.rescaling_factor * float(exc[j*2+0])
            smpole = self.rescaling_factor * float(exc[j*2+1])
            self.multipoles['normal'][harm].append(nmpole)
            self.multipoles['skew'][harm].append(smpole)

    def _read_text(self, text):
        self._init()
        lines = text.splitlines()
        self.currents = []

        for line in lines:
            line = line.strip()
            if not line:
                continue  # empty line
            if line[0] == '#':
                self._process_comment_line(line)
            else:
                self._process_data(line)

        # sort data
        idx = _np.argsort(self.currents)
        self.currents = [self.currents[idx[i]] for i in range(len(idx))]
        for harm in self._harmonics:
            self.multipoles['normal'][harm] = \
                [self.multipoles['normal'][harm][idx[i]]
                 for i in range(len(idx))]
            self.multipoles['skew'][harm] = \
                [self.multipoles['skew'][harm][idx[i]]
                 for i in range(len(idx))]
        self._current_limits = (min(self.currents), max(self.currents))

    def _read_file(self, filename):

        with open(filename) as f:
            text = f.read()
        self._read_text(text)

    def _read_webs(self, label):

        self._fname = label
        text = _web.magnets_excitation_data_read(label)
        self._read_text(text)
